# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
import time

import pytest
from testing_support.fixtures import override_generic_settings
from testing_support.util import conditional_decorator
from testing_support.validators.validate_metric_payload import validate_metric_payload

from newrelic.common.streaming_utils import StreamBuffer
from newrelic.core.agent_streaming import StreamingRpc
from newrelic.core.application import Application
from newrelic.core.config import global_settings
from newrelic.core.infinite_tracing_pb2 import AttributeValue, Span
from newrelic.packages import six

settings = global_settings()

CONDITION_CLS = type(threading.Condition())


@pytest.fixture()
def app():
    app = Application("Python Agent Test (Infinite Tracing)")
    yield app
    # Calling internal_agent_shutdown on an application that is already closed
    # will raise an exception.
    active_session = app._active_session
    try:
        app.internal_agent_shutdown(restart=False)
    except:
        pass
    if active_session and active_session._rpc is not None:
        assert not active_session._rpc.response_processing_thread.is_alive()
        assert not active_session._rpc.channel


@pytest.mark.parametrize(
    "status_code, metrics",
    (
        (
            "UNIMPLEMENTED",
            [
                ("Supportability/InfiniteTracing/Span/gRPC/UNIMPLEMENTED", 1),
                ("Supportability/InfiniteTracing/Span/Response/Error", 1),
            ],
        ),
        (
            "INTERNAL",
            [
                ("Supportability/InfiniteTracing/Span/gRPC/INTERNAL", 1),
                ("Supportability/InfiniteTracing/Span/Response/Error", 1),
            ],
        ),
    ),
)
def test_infinite_tracing_span_streaming(mock_grpc_server, status_code, metrics, monkeypatch, app, batching):
    event = threading.Event()

    class TerminateOnWait(CONDITION_CLS):
        def notify_all(self, *args, **kwargs):
            event.set()
            return super(TerminateOnWait, self).notify_all(*args, **kwargs)

        def wait(self, *args, **kwargs):
            event.set()
            return super(TerminateOnWait, self).wait(*args, **kwargs)

    @staticmethod
    def condition(*args, **kwargs):
        return TerminateOnWait(*args, **kwargs)

    monkeypatch.setattr(StreamingRpc, "condition", condition)

    span = Span(
        intrinsics={"status_code": AttributeValue(string_value=status_code)}, agent_attributes={}, user_attributes={}
    )

    @override_generic_settings(
        settings,
        {
            "distributed_tracing.enabled": True,
            "span_events.enabled": True,
            "infinite_tracing.trace_observer_host": "localhost",
            "infinite_tracing.trace_observer_port": mock_grpc_server,
            "infinite_tracing.ssl": False,
            "infinite_tracing.batching": batching,
        },
    )
    @validate_metric_payload(metrics)
    def _test():
        app.connect_to_data_collector(None)

        app._stats_engine.span_stream.put(span)

        assert event.wait(timeout=5)

        app.harvest(shutdown=True)

    _test()


def test_reconnect_on_failure(monkeypatch, mock_grpc_server, buffer_empty_event, app, batching):
    status_code = "INTERNAL"
    wait_event = threading.Event()
    continue_event = threading.Event()

    class WaitOnWait(CONDITION_CLS):
        def wait(self, *args, **kwargs):
            wait_event.set()
            continue_event.wait()
            return True

    @staticmethod
    def condition(*args, **kwargs):
        return WaitOnWait(*args, **kwargs)

    monkeypatch.setattr(StreamingRpc, "condition", condition)

    terminating_span = Span(
        intrinsics={"status_code": AttributeValue(string_value=status_code)}, agent_attributes={}, user_attributes={}
    )

    span = Span(intrinsics={}, agent_attributes={}, user_attributes={})

    @override_generic_settings(
        settings,
        {
            "distributed_tracing.enabled": True,
            "span_events.enabled": True,
            "infinite_tracing.trace_observer_host": "localhost",
            "infinite_tracing.trace_observer_port": mock_grpc_server,
            "infinite_tracing.ssl": False,
            "infinite_tracing.batching": batching,
        },
    )
    def _test():
        app.connect_to_data_collector(None)

        # Send a span that will trigger a failure
        app._stats_engine.span_stream.put(terminating_span)

        assert wait_event.wait(timeout=5)

        # Send a normal span afterwards
        app._stats_engine.span_stream.put(span)

        buffer_empty_event.clear()

        # Trigger the event so that a reconnect will occur
        continue_event.set()

        # Wait for the stream buffer to empty meaning all spans have been sent.
        assert buffer_empty_event.wait(10)
        app.internal_agent_shutdown(restart=False)

    _test()


def test_agent_restart(app):
    # Get the application connected to the actual 8T endpoint
    app.connect_to_data_collector(None)
    rpc = app._active_session._rpc

    # Store references to the original rpc and threads
    original_rpc = rpc.rpc
    original_thread = rpc.response_processing_thread
    original_span_stream = app._stats_engine.span_stream
    assert original_rpc
    assert rpc.response_processing_thread.is_alive()

    # Force an agent restart
    app.internal_agent_shutdown(restart=True)

    # Wait for connect to complete
    app._connected_event.wait()
    rpc = app._active_session._rpc

    assert not original_thread.is_alive()
    assert rpc.rpc is not original_rpc
    assert app._stats_engine.span_stream is not original_span_stream
    assert rpc.response_processing_thread.is_alive()


def test_disconnect_on_UNIMPLEMENTED(mock_grpc_server, monkeypatch, app, batching):
    event = threading.Event()

    class WaitOnNotify(CONDITION_CLS):
        def notify_all(self, *args, **kwargs):
            event.set()
            return super(WaitOnNotify, self).notify_all(*args, **kwargs)

    @staticmethod
    def condition(*args, **kwargs):
        return WaitOnNotify(*args, **kwargs)

    monkeypatch.setattr(StreamingRpc, "condition", condition)

    terminating_span = Span(
        intrinsics={"status_code": AttributeValue(string_value="UNIMPLEMENTED")},
        agent_attributes={},
        user_attributes={},
    )

    @override_generic_settings(
        settings,
        {
            "distributed_tracing.enabled": True,
            "span_events.enabled": True,
            "infinite_tracing.trace_observer_host": "localhost",
            "infinite_tracing.trace_observer_port": mock_grpc_server,
            "infinite_tracing.ssl": False,
            "infinite_tracing.batching": batching,
        },
    )
    def _test():
        app.connect_to_data_collector(None)

        # Send a span that will trigger disconnect
        app._stats_engine.span_stream.put(terminating_span)

        # Wait for the notify event in close to be called
        assert event.wait(timeout=5)

        # Verify the rpc management thread is killed
        rpc_thread = app._active_session._rpc.response_processing_thread
        rpc_thread.join(timeout=5)
        assert not rpc_thread.is_alive()

    _test()


def test_agent_shutdown():
    # Get the application connected to the actual 8T endpoint
    app = Application("Python Agent Test (Infinite Tracing)")
    app.connect_to_data_collector(None)
    rpc = app._active_session._rpc
    # Store references to the original rpc and threads
    assert rpc.response_processing_thread.is_alive()
    app.internal_agent_shutdown(restart=False)
    assert not rpc.response_processing_thread.is_alive()
    assert not rpc.channel


@pytest.mark.xfail(reason="This test is flaky", strict=False)
def test_no_delay_on_ok(mock_grpc_server, monkeypatch, app, batching):
    wait_event = threading.Event()
    connect_event = threading.Event()

    metrics = [
        ("Supportability/InfiniteTracing/Span/gRPC/OK", 1),
        ("Supportability/InfiniteTracing/Span/Response/Error", None),
    ]

    class SetFlagOnWait(CONDITION_CLS):
        def __init__(self, event, *args, **kwargs):
            super(SetFlagOnWait, self).__init__(*args, **kwargs)
            self.event = event

        def wait(self, *args, **kwargs):
            self.event.set()
            return super(SetFlagOnWait, self).wait(*args, **kwargs)

    @staticmethod
    def condition(*args, **kwargs):
        return SetFlagOnWait(wait_event, *args, **kwargs)

    _create_channel = StreamingRpc.create_channel

    def create_channel(self, *args, **kwargs):
        ret = _create_channel(self, *args, **kwargs)
        connect_event.set()
        return ret

    monkeypatch.setattr(StreamingRpc, "condition", condition)
    monkeypatch.setattr(StreamingRpc, "create_channel", create_channel)

    span = Span(
        intrinsics={"status_code": AttributeValue(string_value="OK")},
        agent_attributes={},
        user_attributes={},
    )

    @override_generic_settings(
        settings,
        {
            "distributed_tracing.enabled": True,
            "span_events.enabled": True,
            "infinite_tracing.trace_observer_host": "localhost",
            "infinite_tracing.trace_observer_port": mock_grpc_server,
            "infinite_tracing.ssl": False,
            "infinite_tracing.batching": batching,
        },
    )
    @validate_metric_payload(metrics)
    def _test():
        def connect_complete():
            connect_event.set()

        app.connect_to_data_collector(connect_complete)

        assert connect_event.wait(timeout=5)
        connect_event.clear()

        # Send a span that will trigger disconnect
        stream_buffer = app._stats_engine.span_stream
        rpc = app._active_session._rpc

        # Put a span that will trigger an OK status code and wait for an attempted
        # reconnect.
        stream_buffer.put(span)
        assert connect_event.wait(timeout=5)
        rpc.close()
        assert not wait_event.is_set()
        app.harvest()

    _test()


@conditional_decorator(
    condition=six.PY2, decorator=pytest.mark.xfail(reason="Test frequently times out on Py2.", strict=False)
)
def test_no_data_loss_on_reconnect(mock_grpc_server, app, buffer_empty_event, batching, spans_processed_event):
    """
    Test for data loss when channel is closed by the server while waiting for more data in a request iterator.

    This is a bug that's caused by the periodic (15 second) disconnects issued by the trace observer. To observe,
    wait long enough in __next__'s notify.wait() call until the server issues a grpc.StatusCode.OK causing a
    disconnect and reconnect. Alternatively in the case of this test, we use a mock server to issue one at the
    appropriate moment rather than waiting for a real trace observer to issue a disconnect.

    While in this state, the very next span placed in the StreamBuffer would wake up the request_iterator for the
    now closed channel (which was waiting in the __next__ function) and be consumed. The channel, being closed,
    would discard the data and finish shutting down. This is now prevented by guards checking if the channel is
    closed before popping any data inside the request iterator, which instead raises a StopIteration.

    Relevant GitHub issue: https://github.com/grpc/grpc/issues/29110
    """

    terminating_span = Span(
        intrinsics={"wait_then_ok": AttributeValue(string_value="OK")}, agent_attributes={}, user_attributes={}
    )

    span = Span(intrinsics={}, agent_attributes={}, user_attributes={})

    @override_generic_settings(
        settings,
        {
            "distributed_tracing.enabled": True,
            "span_events.enabled": True,
            "infinite_tracing.trace_observer_host": "localhost",
            "infinite_tracing.trace_observer_port": mock_grpc_server,
            "infinite_tracing.ssl": False,
            "infinite_tracing.batching": batching,
        },
    )
    def _test():
        # Connect to app and retrieve references to various components
        app.connect_to_data_collector(None)

        stream_buffer = app._stats_engine.span_stream
        rpc = app._active_session._rpc
        request_iterator = rpc.request_iterator

        # Wait until iterator is waiting on spans
        assert buffer_empty_event.wait(timeout=5)
        buffer_empty_event.clear()

        # Send a span that will trigger disconnect
        stream_buffer.put(terminating_span)

        # Wait for spans to be processed by server
        assert spans_processed_event.wait(timeout=5)
        spans_processed_event.clear()

        # Wait for OK status code to close the channel
        start_time = time.time()
        while not (request_iterator._stream and request_iterator._stream.done()):
            assert time.time() - start_time < 15, "Timed out waiting for OK status code."
            time.sleep(0.5)

        # Put new span and wait until buffer has been emptied and either sent or lost
        stream_buffer.put(span)
        assert spans_processed_event.wait(timeout=15), "Data lost in stream buffer iterator."

    _test()


@pytest.mark.parametrize("dropped_spans", [0, 1])
def test_span_supportability_metrics(mock_grpc_server, monkeypatch, app, dropped_spans, batching):
    wait_event = threading.Event()
    continue_event = threading.Event()

    total_spans = 3
    metrics = [
        ("Supportability/InfiniteTracing/Span/Seen", total_spans),
        (
            "Supportability/InfiniteTracing/Span/Sent",
            (total_spans - dropped_spans) or None,
        ),  # Replace 0 with None to indicate metric will not be sent
    ]

    class WaitOnWait(CONDITION_CLS):
        def wait(self, *args, **kwargs):
            wait_event.set()
            ret = super(WaitOnWait, self).wait(*args, **kwargs)
            assert continue_event.wait(timeout=5)
            return ret

    @staticmethod
    def condition(*args, **kwargs):
        return WaitOnWait(*args, **kwargs)

    monkeypatch.setattr(StreamBuffer, "condition", condition)

    span = Span(
        intrinsics={},
        agent_attributes={},
        user_attributes={},
    )

    @override_generic_settings(
        settings,
        {
            "distributed_tracing.enabled": True,
            "span_events.enabled": True,
            "infinite_tracing.trace_observer_host": "localhost",
            "infinite_tracing.trace_observer_port": mock_grpc_server,
            "infinite_tracing.ssl": False,
            "infinite_tracing.batching": batching,
            "infinite_tracing.span_queue_size": total_spans - dropped_spans,
        },
    )
    @validate_metric_payload(metrics)
    def _test():
        app.connect_to_data_collector(None)

        assert wait_event.wait(timeout=5)

        stream_buffer = app._stats_engine.span_stream

        # Put enough spans to overflow buffer
        for _ in range(total_spans):
            stream_buffer.put(span)

        # Harvest all spans simultaneously
        wait_event.clear()
        continue_event.set()
        assert wait_event.wait(timeout=5)
        wait_event.clear()

        app.harvest()

    _test()


@pytest.mark.parametrize("trace_observer_host", ["localhost", None])
@pytest.mark.parametrize("batching", [True, False])
@pytest.mark.parametrize("compression", [True, False])
def test_settings_supportability_metrics(mock_grpc_server, app, trace_observer_host, batching, compression):
    connect_event = threading.Event()

    enabled = bool(trace_observer_host)

    metrics = [
        ("Supportability/InfiniteTracing/gRPC/Batching/enabled", 1 if enabled and batching else None),
        ("Supportability/InfiniteTracing/gRPC/Batching/disabled", 1 if enabled and not batching else None),
        ("Supportability/InfiniteTracing/gRPC/Compression/enabled", 1 if enabled and compression else None),
        ("Supportability/InfiniteTracing/gRPC/Compression/disabled", 1 if enabled and not compression else None),
    ]

    @override_generic_settings(
        settings,
        {
            "distributed_tracing.enabled": True,
            "span_events.enabled": True,
            "infinite_tracing.trace_observer_host": trace_observer_host,
            "infinite_tracing.trace_observer_port": mock_grpc_server,
            "infinite_tracing.ssl": False,
            "infinite_tracing.batching": batching,
            "infinite_tracing.compression": compression,
        },
    )
    @validate_metric_payload(metrics)
    def _test():
        def connect_complete():
            connect_event.set()

        app.connect_to_data_collector(connect_complete)

        assert connect_event.wait(timeout=5)
        connect_event.clear()

        app.harvest()

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import aredis

from testing_support.db_settings import redis_settings

DB_SETTINGS = redis_settings()[0]

strict_redis_client = aredis.StrictRedis(host=DB_SETTINGS['host'], port=DB_SETTINGS['port'], db=0)


IGNORED_METHODS = {
    "cache",
    "connection_pool",
    "execute_command",
    "from_url",
    "hscan_iter",
    "lock",
    "NODES_FLAGS",
    "parse_response",
    "pipeline",
    "register_script",
    "response_callbacks",
    "RESPONSE_CALLBACKS",
    "RESULT_CALLBACKS",
    "sentinel",
    "set_response_callback",
    "transaction",
}

def test_uninstrumented_methods():
    methods = {m for m in dir(strict_redis_client) if not m[0] == "_"}
    is_wrapped = lambda m: hasattr(getattr(strict_redis_client, m), "__wrapped__")
    uninstrumented = {m for m in methods - IGNORED_METHODS if not is_wrapped(m)}

    assert not uninstrumented, "Uninstrumented methods: %s" % sorted(uninstrumented)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from newrelic.api.background_task import background_task
from newrelic.api.function_trace import FunctionTrace

from testing_support.fixtures import validate_tt_parenting
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics

_test_function_trace_default_group_scoped_metrics = [
        ('Function/FunctionTrace', 1)]


@validate_transaction_metrics(
        'test_function_trace:test_function_trace_default_group',
        scoped_metrics=_test_function_trace_default_group_scoped_metrics,
        background_task=True)
@background_task()
def test_function_trace_default_group():
    with FunctionTrace('FunctionTrace'):
        pass


_test_function_trace_valid_override_scoped_metrics = [
        ('Custom/FunctionTrace', 1)]


@validate_transaction_metrics(
        'test_function_trace:test_function_trace_valid_override',
        scoped_metrics=_test_function_trace_valid_override_scoped_metrics,
        background_task=True)
@background_task()
def test_function_trace_valid_override():
    with FunctionTrace('FunctionTrace', group='Custom'):
        pass


_test_function_trace_empty_group_scoped_metrics = [
        ('Function/FunctionTrace', 1)]


@validate_transaction_metrics(
        'test_function_trace:test_function_trace_empty_group',
        scoped_metrics=_test_function_trace_empty_group_scoped_metrics,
        background_task=True)
@background_task()
def test_function_trace_empty_group():
    with FunctionTrace('FunctionTrace', group=''):
        pass


_test_function_trace_leading_slash_on_group_scoped_metrics = [
        ('Function/Group/FunctionTrace', 1)]


@validate_transaction_metrics(
    'test_function_trace:test_function_trace_leading_slash_on_group',
    scoped_metrics=_test_function_trace_leading_slash_on_group_scoped_metrics,
    background_task=True)
@background_task()
def test_function_trace_leading_slash_on_group():
    with FunctionTrace('FunctionTrace', group='/Group'):
        pass


_test_async_trace_overlapping_children_scoped_metrics = [
        ('Function/parent', 1),
        ('Function/child_1', 1),
        ('Function/child_2', 1),
]

_test_async_trace_overlapping_children_parenting = (
        'test_function_trace:test_async_trace_overlapping_children', [
            ('parent', [
                ('child_1', []),
                ('child_2', []),
            ]),
        ]
)


@validate_transaction_metrics(
        'test_function_trace:test_async_trace_overlapping_children',
        scoped_metrics=_test_async_trace_overlapping_children_scoped_metrics,
        background_task=True)
@validate_tt_parenting(_test_async_trace_overlapping_children_parenting)
@background_task()
def test_async_trace_overlapping_children():
    parent_trace = FunctionTrace('parent')

    with parent_trace:
        child_trace_1 = FunctionTrace('child_1', parent=parent_trace)
        child_trace_2 = FunctionTrace('child_2', parent=parent_trace)

        child_trace_1.__enter__()
        child_trace_2.__enter__()
        time.sleep(0.01)
        child_trace_1.__exit__(None, None, None)
        child_trace_2.__exit__(None, None, None)

    assert parent_trace.start_time < child_trace_1.start_time
    assert child_trace_1.start_time < child_trace_2.start_time
    assert child_trace_1.end_time < child_trace_2.end_time
    assert child_trace_2.end_time < parent_trace.end_time


@background_task()
def test_function_trace_settings():
    with FunctionTrace("test_trace") as trace:
        assert trace.settings


def test_function_trace_settings_no_transaction():
    with FunctionTrace("test_trace") as trace:
        assert not trace.settings



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from newrelic.api.database_trace import DatabaseTrace, register_database_client
from newrelic.api.function_trace import FunctionTrace
from newrelic.api.transaction import current_transaction
from newrelic.common.object_names import callable_name
from newrelic.common.object_wrapper import wrap_object
from newrelic.hooks.database_dbapi2 import ConnectionFactory as DBAPI2ConnectionFactory
from newrelic.hooks.database_dbapi2 import ConnectionWrapper as DBAPI2ConnectionWrapper


class ConnectionWrapper(DBAPI2ConnectionWrapper):
    def __enter__(self):
        transaction = current_transaction()
        name = callable_name(self.__wrapped__.__enter__)
        with FunctionTrace(name, source=self.__wrapped__.__enter__):
            cursor = self.__wrapped__.__enter__()

        # The __enter__() method of original connection object returns
        # a new cursor instance for use with 'as' assignment. We need
        # to wrap that in a cursor wrapper otherwise we will not track
        # any queries done via it.

        return self.__cursor_wrapper__(cursor, self._nr_dbapi2_module, self._nr_connect_params, None)

    def __exit__(self, exc, value, tb):
        transaction = current_transaction()
        name = callable_name(self.__wrapped__.__exit__)
        with FunctionTrace(name, source=self.__wrapped__.__exit__):
            if exc is None:
                with DatabaseTrace(
                    "COMMIT", self._nr_dbapi2_module, self._nr_connect_params, source=self.__wrapped__.__exit__
                ):
                    return self.__wrapped__.__exit__(exc, value, tb)
            else:
                with DatabaseTrace(
                    "ROLLBACK", self._nr_dbapi2_module, self._nr_connect_params, source=self.__wrapped__.__exit__
                ):
                    return self.__wrapped__.__exit__(exc, value, tb)


class ConnectionFactory(DBAPI2ConnectionFactory):

    __connection_wrapper__ = ConnectionWrapper


def instance_info(args, kwargs):
    def _bind_params(
        host=None,
        user=None,
        passwd=None,
        db=None,
        port=None,
        unix_socket=None,
        conv=None,
        connect_timeout=None,
        compress=None,
        named_pipe=None,
        init_command=None,
        read_default_file=None,
        read_default_group=None,
        *args,
        **kwargs
    ):
        return (host, port, db, unix_socket, read_default_file, read_default_group)

    params = _bind_params(*args, **kwargs)
    host, port, db, unix_socket, read_default_file, read_default_group = params
    explicit_host = host

    port_path_or_id = None
    if read_default_file or read_default_group:
        host = host or "default"
        port_path_or_id = "unknown"
    elif not host:
        host = "localhost"

    if host == "localhost":
        # precedence: explicit -> cnf (if used) -> env -> 'default'
        port_path_or_id = unix_socket or port_path_or_id or os.getenv("MYSQL_UNIX_PORT", "default")
    elif explicit_host:
        # only reach here if host is explicitly passed in
        port = port and str(port)
        # precedence: explicit -> cnf (if used) -> env -> '3306'
        port_path_or_id = port or port_path_or_id or os.getenv("MYSQL_TCP_PORT", "3306")

    # There is no default database if omitted from the connect params
    # In this case, we should report unknown
    db = db or "unknown"

    return (host, port_path_or_id, db)


def instrument_mysqldb(module):
    register_database_client(
        module,
        database_product="MySQL",
        quoting_style="single+double",
        explain_query="explain",
        explain_stmts=("select",),
        instance_info=instance_info,
    )

    wrap_object(module, "connect", ConnectionFactory, (module,))

    # The connect() function is actually aliased with Connect() and
    # Connection, the later actually being the Connection type object.
    # Instrument Connect(), but don't instrument Connection in case that
    # interferes with direct type usage. If people are using the
    # Connection object directly, they should really be using connect().

    if hasattr(module, "Connect"):
        wrap_object(module, "Connect", ConnectionFactory, (module,))



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys
import uuid

from newrelic.api.function_trace import FunctionTrace
from newrelic.api.time_trace import current_trace
from newrelic.api.transaction import current_transaction
from newrelic.common.object_wrapper import ObjectProxy, wrap_function_wrapper
from newrelic.core.config import global_settings

METHODS_TO_WRAP = ("predict", "fit", "fit_predict", "predict_log_proba", "predict_proba", "transform", "score")
METRIC_SCORERS = (
    "accuracy_score",
    "balanced_accuracy_score",
    "f1_score",
    "precision_score",
    "recall_score",
    "roc_auc_score",
    "r2_score",
)
PY2 = sys.version_info[0] == 2
_logger = logging.getLogger(__name__)


def isnumeric(column):
    import numpy as np

    try:
        column.astype(np.float64)
        return [True] * len(column)
    except:
        pass
    return [False] * len(column)


class PredictReturnTypeProxy(ObjectProxy):
    def __init__(self, wrapped, model_name, training_step):
        super(ObjectProxy, self).__init__(wrapped)
        self._nr_model_name = model_name
        self._nr_training_step = training_step


def _wrap_method_trace(module, class_, method, name=None, group=None):
    def _nr_wrapper_method(wrapped, instance, args, kwargs):
        transaction = current_transaction()
        trace = current_trace()

        if transaction is None:
            return wrapped(*args, **kwargs)

        settings = transaction.settings if transaction.settings is not None else global_settings()

        if settings and not settings.machine_learning.enabled:
            return wrapped(*args, **kwargs)

        wrapped_attr_name = "_nr_wrapped_%s" % method

        # If the method has already been wrapped do not wrap it again. This happens
        # when one class inherits from another and they both implement the method.
        if getattr(trace, wrapped_attr_name, False):
            return wrapped(*args, **kwargs)

        trace = FunctionTrace(name=name, group=group, source=wrapped)

        try:
            # Set the _nr_wrapped attribute to denote that this method is being wrapped.
            setattr(trace, wrapped_attr_name, True)

            with trace:
                return_val = wrapped(*args, **kwargs)
        finally:
            # Set the _nr_wrapped attribute to denote that this method is no longer wrapped.
            setattr(trace, wrapped_attr_name, False)

        # If this is the fit method, increment the training_step counter.
        if method in ("fit", "fit_predict"):
            training_step = getattr(instance, "_nr_wrapped_training_step", -1)
            setattr(instance, "_nr_wrapped_training_step", training_step + 1)

        # If this is the predict method, wrap the return type in an nr type with
        # _nr_wrapped attrs that will attach model info to the data.
        if method in ("predict", "fit_predict"):
            training_step = getattr(instance, "_nr_wrapped_training_step", "Unknown")
            create_prediction_event(transaction, class_, instance, args, kwargs, return_val)
            return PredictReturnTypeProxy(return_val, model_name=class_, training_step=training_step)
        return return_val

    wrap_function_wrapper(module, "%s.%s" % (class_, method), _nr_wrapper_method)


def _calc_prediction_feature_stats(prediction_input, class_, feature_column_names, tags):
    import numpy as np

    # Drop any feature columns that are not numeric since we can't compute stats
    # on non-numeric columns.
    x = np.array(prediction_input)
    isnumeric_features = np.apply_along_axis(isnumeric, 0, x)
    numeric_features = x[isnumeric_features]

    # Drop any feature column names that are not numeric since we can't compute stats
    # on non-numeric columns.
    feature_column_names = feature_column_names[isnumeric_features[0]]

    # Only compute stats for features if we have any feature columns left after dropping
    # non-numeric columns.
    num_cols = len(feature_column_names)
    if num_cols > 0:
        # Boolean selection of numpy array values reshapes the array to a single
        # dimension so we have to reshape it back into a 2D array.
        features = np.reshape(numeric_features, (len(numeric_features) // num_cols, num_cols))
        features = features.astype(dtype=np.float64)

        _record_stats(features, feature_column_names, class_, "Feature", tags)


def _record_stats(data, column_names, class_, column_type, tags):
    import numpy as np

    mean = np.mean(data, axis=0)
    percentile25 = np.percentile(data, q=0.25, axis=0)
    percentile50 = np.percentile(data, q=0.50, axis=0)
    percentile75 = np.percentile(data, q=0.75, axis=0)
    standard_deviation = np.std(data, axis=0)
    _min = np.min(data, axis=0)
    _max = np.max(data, axis=0)
    _count = data.shape[0]

    transaction = current_transaction()

    # Currently record_metric only supports a subset of these stats so we have
    # to upload them one at a time instead of as a dictionary of stats per
    # feature column.
    for index, col_name in enumerate(column_names):
        metric_name = "MLModel/Sklearn/Named/%s/Predict/%s/%s" % (class_, column_type, col_name)

        transaction.record_dimensional_metrics(
            [
                ("%s/%s" % (metric_name, "Mean"), float(mean[index]), tags),
                ("%s/%s" % (metric_name, "Percentile25"), float(percentile25[index]), tags),
                ("%s/%s" % (metric_name, "Percentile50"), float(percentile50[index]), tags),
                ("%s/%s" % (metric_name, "Percentile75"), float(percentile75[index]), tags),
                ("%s/%s" % (metric_name, "StandardDeviation"), float(standard_deviation[index]), tags),
                ("%s/%s" % (metric_name, "Min"), float(_min[index]), tags),
                ("%s/%s" % (metric_name, "Max"), float(_max[index]), tags),
                ("%s/%s" % (metric_name, "Count"), _count, tags),
            ]
        )


def _calc_prediction_label_stats(labels, class_, label_column_names, tags):
    import numpy as np

    labels = np.array(labels, dtype=np.float64)
    _record_stats(labels, label_column_names, class_, "Label", tags)


def _get_label_names(user_defined_label_names, prediction_array):
    import numpy as np

    if user_defined_label_names is None:
        return np.array(range(prediction_array.shape[1]))
    if user_defined_label_names and len(user_defined_label_names) != prediction_array.shape[1]:
        _logger.warning(
            "The number of label names passed to the ml_model wrapper function is not equal to the number of predictions in the data set. Please supply the correct number of label names."
        )
        return np.array(range(prediction_array.shape[1]))
    else:
        return user_defined_label_names


def find_type_category(data_set, row_index, column_index):
    # If pandas DataFrame, return type of column.
    pd = sys.modules.get("pandas", None)
    if pd and isinstance(data_set, pd.DataFrame):
        value_type = data_set.iloc[:, column_index].dtype.name
        if value_type == "category":
            return "categorical"
        categorized_value_type = categorize_data_type(value_type)
        return categorized_value_type
    # If it's not a pandas DataFrame then it is a list or numpy array.
    python_type = str(type(data_set[column_index][row_index]))
    return categorize_data_type(python_type)


def categorize_data_type(python_type):
    if "int" in python_type or "float" in python_type or "complex" in python_type:
        return "numeric"
    if "bool" in python_type:
        return "bool"
    if "str" in python_type or "unicode" in python_type:
        return "str"
    else:
        return python_type


def _get_feature_column_names(user_provided_feature_names, features):
    import numpy as np

    num_feature_columns = np.array(features).shape[1]

    # If the user provided feature names are the correct size, return the user provided feature
    # names.
    if user_provided_feature_names and len(user_provided_feature_names) == num_feature_columns:
        return np.array(user_provided_feature_names)

    # If the user provided feature names aren't the correct size, log a warning and do not use the user provided feature names.
    if user_provided_feature_names:
        _logger.warning(
            "The number of feature names passed to the ml_model wrapper function is not equal to the number of columns in the data set. Please supply the correct number of feature names."
        )

    # If the user doesn't provide the feature names or they were provided but the size was incorrect and the features are a pandas data frame, return the column names from the pandas data frame.
    pd = sys.modules.get("pandas", None)
    if pd and isinstance(features, pd.DataFrame):
        return features.columns

    # If the user doesn't provide the feature names or they were provided but the size was incorrect and the features are not a pandas data frame, return the column indexes as the feature names.
    return np.array(range(num_feature_columns))


def bind_predict(X, *args, **kwargs):
    return X


def create_prediction_event(transaction, class_, instance, args, kwargs, return_val):
    import numpy as np

    data_set = bind_predict(*args, **kwargs)
    model_name = getattr(instance, "_nr_wrapped_name", class_)
    model_version = getattr(instance, "_nr_wrapped_version", "0.0.0")
    user_provided_feature_names = getattr(instance, "_nr_wrapped_feature_names", None)
    label_names = getattr(instance, "_nr_wrapped_label_names", None)
    metadata = getattr(instance, "_nr_wrapped_metadata", {})
    settings = transaction.settings if transaction.settings is not None else global_settings()

    prediction_id = uuid.uuid4()

    labels = []
    if return_val is not None:
        if not hasattr(return_val, "__iter__"):
            labels = np.array([return_val])
        else:
            labels = np.array(return_val)
        if len(labels.shape) == 1:
            labels = np.reshape(labels, (len(labels) // 1, 1))

        label_names_list = _get_label_names(label_names, labels)
        _calc_prediction_label_stats(
            labels,
            class_,
            label_names_list,
            tags={
                "prediction_id": prediction_id,
                "model_version": model_version,
                # The following are used for entity synthesis.
                "modelName": model_name,
            },
        )

    final_feature_names = _get_feature_column_names(user_provided_feature_names, data_set)
    np_casted_data_set = np.array(data_set)
    _calc_prediction_feature_stats(
        data_set,
        class_,
        final_feature_names,
        tags={
            "prediction_id": prediction_id,
            "model_version": model_version,
            # The following are used for entity synthesis.
            "modelName": model_name,
        },
    )
    features, predictions = np_casted_data_set.shape
    for prediction_index, prediction in enumerate(np_casted_data_set):
        inference_id = uuid.uuid4()

        event = {
            "inference_id": inference_id,
            "prediction_id": prediction_id,
            "model_version": model_version,
            "new_relic_data_schema_version": 2,
            # The following are used for entity synthesis.
            "modelName": model_name,
        }
        if metadata and isinstance(metadata, dict):
            event.update(metadata)
        # Don't include the raw value when inference_event_value is disabled.
        if settings and settings.machine_learning and settings.machine_learning.inference_events_value.enabled:
            event.update(
                {
                    "feature.%s" % str(final_feature_names[feature_col_index]): value
                    for feature_col_index, value in enumerate(prediction)
                }
            )
            event.update(
                {
                    "label.%s" % str(label_names_list[index]): str(value)
                    for index, value in enumerate(labels[prediction_index])
                }
            )
        transaction.record_ml_event("InferenceData", event)


def _nr_instrument_model(module, model_class):
    for method_name in METHODS_TO_WRAP:
        if hasattr(getattr(module, model_class), method_name):
            # Function/MLModel/Sklearn/Named/<class name>.<method name>
            name = "MLModel/Sklearn/Named/%s.%s" % (model_class, method_name)
            _wrap_method_trace(module, model_class, method_name, name=name)


def _instrument_sklearn_models(module, model_classes):
    for model_cls in model_classes:
        if hasattr(module, model_cls):
            _nr_instrument_model(module, model_cls)


def _bind_scorer(y_true, y_pred, *args, **kwargs):
    return y_true, y_pred, args, kwargs


def wrap_metric_scorer(wrapped, instance, args, kwargs):
    transaction = current_transaction()
    # If there is no transaction, do not wrap anything.
    if not transaction:
        return wrapped(*args, **kwargs)

    settings = transaction.settings if transaction.settings is not None else global_settings()

    if settings and not settings.machine_learning.enabled:
        return wrapped(*args, **kwargs)

    score = wrapped(*args, **kwargs)

    y_true, y_pred, args, kwargs = _bind_scorer(*args, **kwargs)
    model_name = "Unknown"
    training_step = "Unknown"
    if hasattr(y_pred, "_nr_model_name"):
        model_name = y_pred._nr_model_name
    if hasattr(y_pred, "_nr_training_step"):
        training_step = y_pred._nr_training_step
    # Attribute values must be int, float, str, or boolean. If it's not one of these
    # types and an iterable add the values as separate attributes.
    if not isinstance(score, (str, int, float, bool)):
        if hasattr(score, "__iter__"):
            for i, s in enumerate(score):
                transaction._add_agent_attribute(
                    "%s/TrainingStep/%s/%s[%s]" % (model_name, training_step, wrapped.__name__, i), s
                )
    else:
        transaction._add_agent_attribute("%s/TrainingStep/%s/%s" % (model_name, training_step, wrapped.__name__), score)
    return score


def instrument_sklearn_tree_models(module):
    model_classes = (
        "DecisionTreeClassifier",
        "DecisionTreeRegressor",
        "ExtraTreeClassifier",
        "ExtraTreeRegressor",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_ensemble_bagging_models(module):
    model_classes = (
        "BaggingClassifier",
        "BaggingRegressor",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_ensemble_forest_models(module):
    model_classes = (
        "ExtraTreesClassifier",
        "ExtraTreesRegressor",
        "RandomForestClassifier",
        "RandomForestRegressor",
        "RandomTreesEmbedding",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_ensemble_iforest_models(module):
    model_classes = ("IsolationForest",)
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_ensemble_weight_boosting_models(module):
    model_classes = (
        "AdaBoostClassifier",
        "AdaBoostRegressor",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_ensemble_gradient_boosting_models(module):
    model_classes = (
        "GradientBoostingClassifier",
        "GradientBoostingRegressor",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_ensemble_voting_models(module):
    model_classes = (
        "VotingClassifier",
        "VotingRegressor",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_ensemble_stacking_models(module):
    module_classes = (
        "StackingClassifier",
        "StackingRegressor",
    )
    _instrument_sklearn_models(module, module_classes)


def instrument_sklearn_ensemble_hist_models(module):
    model_classes = (
        "HistGradientBoostingClassifier",
        "HistGradientBoostingRegressor",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_linear_coordinate_descent_models(module):
    model_classes = (
        "Lasso",
        "LassoCV",
        "ElasticNet",
        "ElasticNetCV",
        "MultiTaskLasso",
        "MultiTaskLassoCV",
        "MultiTaskElasticNet",
        "MultiTaskElasticNetCV",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_compose_models(module):
    model_classes = (
        "ColumnTransformer",
        "TransformedTargetRegressor",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_covariance_shrunk_models(module):
    model_classes = (
        "ShrunkCovariance",
        "LedoitWolf",
        "OAS",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_cross_decomposition_models(module):
    model_classes = (
        "PLSRegression",
        "PLSSVD",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_covariance_graph_models(module):
    model_classes = (
        "GraphicalLasso",
        "GraphicalLassoCV",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_discriminant_analysis_models(module):
    model_classes = (
        "LinearDiscriminantAnalysis",
        "QuadraticDiscriminantAnalysis",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_covariance_models(module):
    model_classes = (
        "EmpiricalCovariance",
        "MinCovDet",
        "EllipticEnvelope",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_gaussian_process_models(module):
    model_classes = (
        "GaussianProcessClassifier",
        "GaussianProcessRegressor",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_dummy_models(module):
    model_classes = (
        "DummyClassifier",
        "DummyRegressor",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_feature_selection_rfe_models(module):
    model_classes = (
        "RFE",
        "RFECV",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_kernel_ridge_models(module):
    model_classes = ("KernelRidge",)
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_calibration_models(module):
    model_classes = ("CalibratedClassifierCV",)
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_cluster_models(module):
    model_classes = (
        "AffinityPropagation",
        "Birch",
        "DBSCAN",
        "MeanShift",
        "OPTICS",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_linear_least_angle_models(module):
    model_classes = (
        "Lars",
        "LarsCV",
        "LassoLars",
        "LassoLarsCV",
        "LassoLarsIC",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_feature_selection_models(module):
    model_classes = (
        "VarianceThreshold",
        "SelectFromModel",
        "SequentialFeatureSelector",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_cluster_agglomerative_models(module):
    model_classes = (
        "AgglomerativeClustering",
        "FeatureAgglomeration",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_linear_GLM_models(module):
    model_classes = (
        "PoissonRegressor",
        "GammaRegressor",
        "TweedieRegressor",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_cluster_clustering_models(module):
    model_classes = (
        "SpectralBiclustering",
        "SpectralCoclustering",
        "SpectralClustering",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_linear_stochastic_gradient_models(module):
    model_classes = (
        "SGDClassifier",
        "SGDRegressor",
        "SGDOneClassSVM",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_linear_ridge_models(module):
    model_classes = (
        "Ridge",
        "RidgeCV",
        "RidgeClassifier",
        "RidgeClassifierCV",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_linear_logistic_models(module):
    model_classes = (
        "LogisticRegression",
        "LogisticRegressionCV",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_linear_OMP_models(module):
    model_classes = (
        "OrthogonalMatchingPursuit",
        "OrthogonalMatchingPursuitCV",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_linear_passive_aggressive_models(module):
    model_classes = (
        "PassiveAggressiveClassifier",
        "PassiveAggressiveRegressor",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_linear_bayes_models(module):
    model_classes = (
        "ARDRegression",
        "BayesianRidge",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_linear_models(module):
    model_classes = (
        "HuberRegressor",
        "LinearRegression",
        "Perceptron",
        "QuantileRegressor",
        "TheilSenRegressor",
        "RANSACRegressor",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_cluster_kmeans_models(module):
    model_classes = (
        "BisectingKMeans",
        "KMeans",
        "MiniBatchKMeans",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_multiclass_models(module):
    model_classes = (
        "OneVsRestClassifier",
        "OneVsOneClassifier",
        "OutputCodeClassifier",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_multioutput_models(module):
    model_classes = (
        "MultiOutputEstimator",
        "MultiOutputClassifier",
        "ClassifierChain",
        "RegressorChain",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_naive_bayes_models(module):
    model_classes = (
        "GaussianNB",
        "MultinomialNB",
        "ComplementNB",
        "BernoulliNB",
        "CategoricalNB",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_model_selection_models(module):
    model_classes = (
        "GridSearchCV",
        "RandomizedSearchCV",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_mixture_models(module):
    model_classes = (
        "GaussianMixture",
        "BayesianGaussianMixture",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_neural_network_models(module):
    model_classes = (
        "BernoulliRBM",
        "MLPClassifier",
        "MLPRegressor",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_neighbors_KRadius_models(module):
    model_classes = (
        "KNeighborsClassifier",
        "RadiusNeighborsClassifier",
        "KNeighborsTransformer",
        "RadiusNeighborsTransformer",
        "KNeighborsRegressor",
        "RadiusNeighborsRegressor",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_svm_models(module):
    model_classes = (
        "LinearSVC",
        "LinearSVR",
        "SVC",
        "NuSVC",
        "SVR",
        "NuSVR",
        "OneClassSVM",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_semi_supervised_models(module):
    model_classes = (
        "LabelPropagation",
        "LabelSpreading",
        "SelfTrainingClassifier",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_pipeline_models(module):
    model_classes = (
        "Pipeline",
        "FeatureUnion",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_neighbors_models(module):
    model_classes = (
        "KernelDensity",
        "LocalOutlierFactor",
        "NeighborhoodComponentsAnalysis",
        "NearestCentroid",
        "NearestNeighbors",
    )
    _instrument_sklearn_models(module, model_classes)


def instrument_sklearn_metrics(module):
    for scorer in METRIC_SCORERS:
        if hasattr(module, scorer):
            wrap_function_wrapper(module, scorer, wrap_metric_scorer)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pymssql

from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_database_trace_inputs import validate_database_trace_inputs

from testing_support.db_settings import mssql_settings

from newrelic.api.background_task import background_task

DB_SETTINGS = mssql_settings()[0]
TABLE_NAME = "datastore_pymssql_" + DB_SETTINGS["namespace"]
PROCEDURE_NAME = "hello_" + DB_SETTINGS["namespace"]


def execute_db_calls_with_cursor(cursor):
    cursor.execute("""drop table if exists %s""" % TABLE_NAME)

    cursor.execute("""create table %s """ % TABLE_NAME + """(a integer, b real, c text)""")

    cursor.executemany(
        """insert into %s """ % TABLE_NAME + """values (%s, %s, %s)""",
        [(1, 1.0, "1.0"), (2, 2.2, "2.2"), (3, 3.3, "3.3")],
    )

    cursor.execute("""select * from %s""" % TABLE_NAME)

    for row in cursor:
        pass

    cursor.execute("""update %s""" % TABLE_NAME + """ set a=%s, b=%s, """ """c=%s where a=%s""", (4, 4.0, "4.0", 1))

    cursor.execute("""delete from %s where a=2""" % TABLE_NAME)
    cursor.execute("""drop procedure if exists %s""" % PROCEDURE_NAME)
    cursor.execute(
        """CREATE PROCEDURE %s AS
                      BEGIN
                        SELECT 'Hello World!';
                      END"""
        % PROCEDURE_NAME
    )

    cursor.callproc(PROCEDURE_NAME)


_test_scoped_metrics = [
    ("Function/pymssql._pymssql:connect", 1),
    ("Datastore/statement/MSSQL/%s/select" % TABLE_NAME, 1),
    ("Datastore/statement/MSSQL/%s/insert" % TABLE_NAME, 1),
    ("Datastore/statement/MSSQL/%s/update" % TABLE_NAME, 1),
    ("Datastore/statement/MSSQL/%s/delete" % TABLE_NAME, 1),
    ("Datastore/operation/MSSQL/drop", 2),
    ("Datastore/operation/MSSQL/create", 2),
    ("Datastore/statement/MSSQL/%s/call" % PROCEDURE_NAME, 1),
    ("Datastore/operation/MSSQL/commit", 2),
    ("Datastore/operation/MSSQL/rollback", 1),
]

_test_rollup_metrics = [
    ("Datastore/all", 13),
    ("Datastore/allOther", 13),
    ("Datastore/MSSQL/all", 13),
    ("Datastore/MSSQL/allOther", 13),
    ("Datastore/statement/MSSQL/%s/select" % TABLE_NAME, 1),
    ("Datastore/statement/MSSQL/%s/insert" % TABLE_NAME, 1),
    ("Datastore/statement/MSSQL/%s/update" % TABLE_NAME, 1),
    ("Datastore/statement/MSSQL/%s/delete" % TABLE_NAME, 1),
    ("Datastore/operation/MSSQL/select", 1),
    ("Datastore/operation/MSSQL/insert", 1),
    ("Datastore/operation/MSSQL/update", 1),
    ("Datastore/operation/MSSQL/delete", 1),
    ("Datastore/statement/MSSQL/%s/call" % PROCEDURE_NAME, 1),
    ("Datastore/operation/MSSQL/call", 1),
    ("Datastore/operation/MSSQL/drop", 2),
    ("Datastore/operation/MSSQL/create", 2),
    ("Datastore/operation/MSSQL/commit", 2),
    ("Datastore/operation/MSSQL/rollback", 1),
]


@validate_transaction_metrics(
    "test_database:test_execute_via_cursor_context_manager",
    scoped_metrics=_test_scoped_metrics,
    rollup_metrics=_test_rollup_metrics,
    background_task=True,
)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@background_task()
def test_execute_via_cursor_context_manager():
    connection = pymssql.connect(
        user=DB_SETTINGS["user"], password=DB_SETTINGS["password"], host=DB_SETTINGS["host"], port=DB_SETTINGS["port"]
    )

    with connection:
        cursor = connection.cursor()

        with cursor:
            execute_db_calls_with_cursor(cursor)

        connection.commit()
        connection.rollback()
        connection.commit()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_code_level_metrics import (
    validate_code_level_metrics,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)


def target_application():
    # We need to delay Flask application creation because of ordering
    # issues whereby the agent needs to be initialised before Flask is
    # imported and the routes configured. Normally pytest only runs the
    # global fixture which will initialise the agent after each test
    # file is imported, which is too late. We also can't do application
    # creation within a function as we will then get view handler
    # functions are different between Python 2 and 3, with the latter
    # showing <local> scope in path.

    from _test_compress import _test_application

    return _test_application


_test_application_index_scoped_metrics = [
    ("Function/flask.app:Flask.wsgi_app", 1),
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/_test_compress:index_page", 1),
    ("Function/werkzeug.wsgi:ClosingIterator.close", 1),
]


@validate_code_level_metrics("_test_compress", "index_page")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics("_test_compress:index_page", scoped_metrics=_test_application_index_scoped_metrics)
def test_compress_middleware():
    application = target_application()
    response = application.get("/compress")
    response.mustcontain(500 * "X")


_test_html_insertion_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "browser_monitoring.content_type": ["text/html"],
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion_flask_middleware():
    application = target_application()
    headers = {"Accept-Encoding": "gzip"}
    response = application.get("/html_insertion", headers=headers, status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain("NREUM HEADER", "NREUM.info")


@override_application_settings(_test_html_insertion_settings)
def test_html_inserted_for_html_served_from_file():
    application = target_application()
    headers = {"Accept-Encoding": "gzip"}
    response = application.get("/html_served_from_file", headers=headers, status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain("NREUM HEADER", "NREUM.info")


_test_html_insertion_manual_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "browser_monitoring.content_type": ["text/html"],
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_manual_settings)
def test_html_insertion_manual_flask_middleware():
    application = target_application()
    response = application.get("/html_insertion_manual", status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion_unnamed_attachment_header_flask_middleware():
    application = target_application()
    response = application.get("/html_insertion_unnamed_attachment_header", status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion_named_attachment_header_flask_middleware():
    application = target_application()
    response = application.get("/html_insertion_named_attachment_header", status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])


@override_application_settings(_test_html_insertion_settings)
def test_text_served_from_file():
    application = target_application()
    response = application.get("/text_served_from_file", status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])


@override_application_settings(_test_html_insertion_settings)
def test_empty_content_type():
    application = target_application()
    response = application.get("/empty_content_type")

    # Make sure agent can handle content type of ''

    assert response.headers["Content-Type"] == ""
    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import threading
import traceback

import pytest
from testing_support.fixtures import (
    override_application_settings,
    reset_core_stats_engine,
)
from testing_support.validators.validate_error_event_attributes import (
    validate_error_event_attributes,
)
from testing_support.validators.validate_error_event_attributes_outside_transaction import (
    validate_error_event_attributes_outside_transaction,
)
from testing_support.validators.validate_error_trace_attributes import (
    validate_error_trace_attributes,
)
from testing_support.validators.validate_error_trace_attributes_outside_transaction import (
    validate_error_trace_attributes_outside_transaction,
)

from newrelic.api.application import application_instance as application
from newrelic.api.background_task import background_task
from newrelic.api.settings import set_error_group_callback
from newrelic.api.time_trace import notice_error
from newrelic.api.transaction import current_transaction
from newrelic.api.web_transaction import web_transaction
from newrelic.common.object_names import callable_name

_callback_called = threading.Event()
_truncated_value = "A" * 300


def error_group_callback(exc, data):
    _callback_called.set()

    if isinstance(exc, ValueError):
        return "value"
    elif isinstance(exc, ZeroDivisionError):
        return _truncated_value
    elif isinstance(exc, IndexError):
        return []
    elif isinstance(exc, LookupError):
        return 123
    elif isinstance(exc, TypeError):
        return ""


def test_clear_error_group_callback():
    settings = application().settings
    set_error_group_callback(lambda x, y: None)
    assert settings.error_collector.error_group_callback is not None, "Failed to set callback."
    set_error_group_callback(None)
    assert settings.error_collector.error_group_callback is None, "Failed to clear callback."


@pytest.mark.parametrize(
    "callback,accepted", [(error_group_callback, True), (lambda x, y: None, True), (None, False), ("string", False)]
)
def test_set_error_group_callback(callback, accepted):
    try:
        set_error_group_callback(callback)
        settings = application().settings
        if accepted:
            assert settings.error_collector.error_group_callback is not None, "Failed to set callback."
        else:
            assert settings.error_collector.error_group_callback is None, "Accepted bad callback."
    finally:
        set_error_group_callback(None)


@pytest.mark.parametrize(
    "exc_class,group_name,high_security",
    [
        (ValueError, "value", False),
        (ValueError, "value", True),
        (TypeError, None, False),
        (RuntimeError, None, False),
        (IndexError, None, False),
        (LookupError, None, False),
        (ZeroDivisionError, _truncated_value[:255], False),
    ],
    ids=("standard", "high-security", "empty-string", "None-value", "list-type", "int-type", "truncated-value"),
)
@reset_core_stats_engine()
def test_error_group_name_callback(exc_class, group_name, high_security):
    _callback_called.clear()

    if group_name is not None:
        exact = {"user": {}, "intrinsic": {}, "agent": {"error.group.name": group_name}}
        forgone = None
    else:
        exact = None
        forgone = {"user": [], "intrinsic": [], "agent": ["error.group.name"]}

    @validate_error_trace_attributes(callable_name(exc_class), forgone_params=forgone, exact_attrs=exact)
    @validate_error_event_attributes(forgone_params=forgone, exact_attrs=exact)
    @override_application_settings({"high_security": high_security})
    @background_task()
    def _test():

        try:
            raise exc_class()
        except Exception:
            notice_error()

        assert _callback_called.is_set()

    try:
        set_error_group_callback(error_group_callback)
        _test()
    finally:
        set_error_group_callback(None)


@pytest.mark.parametrize(
    "exc_class,group_name,high_security",
    [
        (ValueError, "value", False),
        (ValueError, "value", True),
        (TypeError, None, False),
        (RuntimeError, None, False),
        (IndexError, None, False),
        (LookupError, None, False),
        (ZeroDivisionError, _truncated_value[:255], False),
    ],
    ids=("standard", "high-security", "empty-string", "None-value", "list-type", "int-type", "truncated-value"),
)
@reset_core_stats_engine()
def test_error_group_name_callback_outside_transaction(exc_class, group_name, high_security):
    _callback_called.clear()

    if group_name is not None:
        exact = {"user": {}, "intrinsic": {}, "agent": {"error.group.name": group_name}}
        forgone = None
    else:
        exact = None
        forgone = {"user": [], "intrinsic": [], "agent": ["error.group.name"]}

    @validate_error_trace_attributes_outside_transaction(
        callable_name(exc_class), forgone_params=forgone, exact_attrs=exact
    )
    @validate_error_event_attributes_outside_transaction(forgone_params=forgone, exact_attrs=exact)
    @override_application_settings({"high_security": high_security})
    def _test():
        try:
            raise exc_class()
        except Exception:
            app = application()
            notice_error(application=app)

        assert _callback_called.is_set()

    try:
        set_error_group_callback(error_group_callback)
        _test()
    finally:
        set_error_group_callback(None)


@pytest.mark.parametrize(
    "transaction_decorator",
    [
        background_task(name="TestBackgroundTask"),
        web_transaction(
            name="TestWebTransaction",
            host="localhost",
            port=1234,
            request_method="GET",
            request_path="/",
            headers=[],
        ),
        None,
    ],
    ids=("background_task", "web_transation", "outside_transaction"),
)
@reset_core_stats_engine()
def test_error_group_name_callback_attributes(transaction_decorator):
    callback_errors = []
    _data = []

    def callback(error, data):
        def _callback():
            import types

            _data.append(data)
            txn = current_transaction()

            # Standard attributes
            assert isinstance(error, Exception)
            assert isinstance(data["traceback"], types.TracebackType)
            assert data["error.class"] is type(error)
            assert data["error.message"] == "text"
            assert data["error.expected"] is False

            # All attributes should always be included, but set to None when not relevant.
            if txn is None:  # Outside transaction
                assert data["transactionName"] is None
                assert data["custom_params"] == {"notice_error_attribute": 1}
                assert data["response.status"] is None
                assert data["request.method"] is None
                assert data["request.uri"] is None
            elif txn.background_task:  # Background task
                assert data["transactionName"] == "TestBackgroundTask"
                assert data["custom_params"] == {"notice_error_attribute": 1, "txn_attribute": 2}
                assert data["response.status"] is None
                assert data["request.method"] is None
                assert data["request.uri"] is None
            else:  # Web transaction
                assert data["transactionName"] == "TestWebTransaction"
                assert data["custom_params"] == {"notice_error_attribute": 1, "txn_attribute": 2}
                assert data["response.status"] == 200
                assert data["request.method"] == "GET"
                assert data["request.uri"] == "/"

        try:
            _callback()
        except Exception:
            callback_errors.append(sys.exc_info())
            raise

    def _test():
        try:
            txn = current_transaction()
            if txn:
                txn.add_custom_attribute("txn_attribute", 2)
                if not txn.background_task:
                    txn.process_response(200, [])
            raise Exception("text")
        except Exception:
            app = application() if transaction_decorator is None else None  # Only set outside transaction
            notice_error(application=app, attributes={"notice_error_attribute": 1})

        assert not callback_errors, "Callback inputs failed to validate.\nerror: %s\ndata: %s" % (
            traceback.format_exception(*callback_errors[0]),
            str(_data[0]),
        )

    if transaction_decorator is not None:
        _test = transaction_decorator(_test)  # Manually decorate test function

    try:
        set_error_group_callback(callback)
        _test()
    finally:
        set_error_group_callback(None)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

INI_FILE_EMPTY = b"""
[newrelic]
"""


INI_FILE_INFINITE_TRACING = b"""
[newrelic]
infinite_tracing.trace_observer_host = y
infinite_tracing.trace_observer_port = 1234
infinite_tracing.span_queue_size = 2000
"""


# Tests for loading settings and testing for values precedence
@pytest.mark.parametrize(
    "ini,env,expected_host,log",
    (
        (INI_FILE_EMPTY, {}, None, None),
        (INI_FILE_EMPTY, {"NEW_RELIC_INFINITE_TRACING_TRACE_OBSERVER_HOST": "x://host:443/path"}, "host", "WARNING"),
        (INI_FILE_EMPTY, {"NEW_RELIC_INFINITE_TRACING_TRACE_OBSERVER_HOST": "who/knows"}, None, "ERROR"),
        (INI_FILE_EMPTY, {"NEW_RELIC_INFINITE_TRACING_TRACE_OBSERVER_HOST": ":shrug:"}, None, "ERROR"),
        (INI_FILE_EMPTY, {"NEW_RELIC_INFINITE_TRACING_TRACE_OBSERVER_HOST": "x"}, "x", None),
        (INI_FILE_INFINITE_TRACING, {"NEW_RELIC_INFINITE_TRACING_TRACE_OBSERVER_HOST": "x"}, "y", None),
    ),
)
def test_infinite_tracing_host(ini, env, expected_host, log, global_settings, caplog):

    settings = global_settings()
    assert settings.infinite_tracing.trace_observer_host == expected_host

    if log:
        records = caplog.get_records("setup")
        assert sum(1 for record in records if record.levelname == log) == 1
    else:
        assert not caplog.get_records("setup")


@pytest.mark.parametrize(
    "ini,env,expected_port",
    (
        (INI_FILE_EMPTY, {}, 443),
        (INI_FILE_EMPTY, {"NEW_RELIC_INFINITE_TRACING_TRACE_OBSERVER_PORT": "6789"}, 6789),
        (INI_FILE_INFINITE_TRACING, {"NEW_RELIC_INFINITE_TRACING_TRACE_OBSERVER_PORT": "6789"}, 1234),
    ),
)
def test_infinite_tracing_port(ini, env, expected_port, global_settings):

    settings = global_settings()
    assert settings.infinite_tracing.trace_observer_port == expected_port


# Tests for loading Infinite Tracing span queue size setting
# and testing values precedence
@pytest.mark.parametrize(
    "ini,env,expected_size",
    (
        (INI_FILE_EMPTY, {}, 10000),
        (INI_FILE_EMPTY, {"NEW_RELIC_INFINITE_TRACING_SPAN_QUEUE_SIZE": "invalid"}, 10000),
        (INI_FILE_EMPTY, {"NEW_RELIC_INFINITE_TRACING_SPAN_QUEUE_SIZE": "5000"}, 5000),
        (INI_FILE_INFINITE_TRACING, {"NEW_RELIC_INFINITE_TRACING_SPAN_QUEUE_SIZE": "3000"}, 2000),
    ),
)
def test_infinite_tracing_span_queue_size(ini, env, expected_size, global_settings):

    settings = global_settings()
    assert settings.infinite_tracing.span_queue_size == expected_size



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


IGNORED_METHODS = {
    "address",
    "auto_close_connection_pool",
    "channels",
    "client_tracking_off",
    "client_tracking_on",
    "client_no_touch",
    "close",
    "closed",
    "connection_pool",
    "connection",
    "db",
    "encoding",
    "execute_command",
    "execute",
    "from_url",
    "get_connection_kwargs",
    "get_encoder",
    "get_retry",
    "hscan_iter",
    "ihscan",
    "in_pubsub",
    "in_transaction",
    "initialize",
    "iscan",
    "isscan",
    "izscan",
    "load_external_module",
    "lock",
    "multi_exec",
    "parse_response",
    "patterns",
    "pipeline",
    "publish_json",
    "register_script",
    "response_callbacks",
    "RESPONSE_CALLBACKS",
    "sentinel",
    "SET_IF_EXIST",
    "SET_IF_NOT_EXIST",
    "set_response_callback",
    "set_retry",
    "SHUTDOWN_NOSAVE",
    "SHUTDOWN_SAVE",
    "single_connection_client",
    "transaction",
    "wait_closed",
    "xinfo",
    "ZSET_AGGREGATE_MAX",
    "ZSET_AGGREGATE_MIN",
    "ZSET_AGGREGATE_SUM",
    "ZSET_EXCLUDE_BOTH",
    "ZSET_EXCLUDE_MAX",
    "ZSET_EXCLUDE_MIN",
    "ZSET_IF_EXIST",
    "ZSET_IF_NOT_EXIST",
}

REDIS_MODULES = {
    "bf",
    "cf",
    "cms",
    "ft",
    "graph",
    "json",
    "tdigest",
    "topk",
    "ts",
}

IGNORED_METHODS |= REDIS_MODULES


def test_uninstrumented_methods(client):
    methods = {m for m in dir(client) if not m[0] == "_"}
    is_wrapped = lambda m: hasattr(getattr(client, m), "__wrapped__")
    uninstrumented = {m for m in methods - IGNORED_METHODS if not is_wrapped(m)}

    assert not uninstrumented, "Uninstrumented methods: %s" % sorted(uninstrumented)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.common.object_wrapper import transient_function_wrapper


def validate_external_node_params(params=[], forgone_params=[]):
    """
    Validate the parameters on the external node.

    params: a list of tuples
    forgone_params: a flat list
    """
    @transient_function_wrapper('newrelic.api.external_trace',
            'ExternalTrace.process_response_headers')
    def _validate_external_node_params(wrapped, instance, args, kwargs):
        result = wrapped(*args, **kwargs)

        # This is only validating that logic to extract cross process
        # header and update params in ExternalTrace is succeeding. This
        # is actually done after the ExternalTrace __exit__() is called
        # with the ExternalNode only being updated by virtue of the
        # original params dictionary being aliased rather than copied.
        # So isn't strictly validating that params ended up in the actual
        # ExternalNode in the transaction trace.

        for name, value in params:
            assert instance.params[name] == value

        for name in forgone_params:
            assert name not in instance.params

        return result

    return _validate_external_node_params



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import psycopg2
import psycopg2.extras
import pytest

from testing_support.fixtures import (
    validate_stats_engine_explain_plan_output_is_none,
    override_application_settings)
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_transaction_errors import validate_transaction_errors
from testing_support.validators.validate_database_trace_inputs import validate_database_trace_inputs
from testing_support.validators.validate_transaction_slow_sql_count import (
    validate_transaction_slow_sql_count)
from testing_support.util import instance_hostname
from testing_support.db_settings import postgresql_settings

DB_SETTINGS = postgresql_settings()[0]

from newrelic.api.background_task import background_task


# Settings

_enable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': True,
}
_disable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': False,
}

# Metrics

_base_scoped_metrics = (
        ('Datastore/statement/Postgres/%s/select' % DB_SETTINGS['table_name'], 1),
        ('Datastore/statement/Postgres/%s/insert' % DB_SETTINGS['table_name'], 1),
        ('Datastore/operation/Postgres/drop', 1),
        ('Datastore/operation/Postgres/create', 1)
)

_base_rollup_metrics = (
        ('Datastore/all', 5),
        ('Datastore/allOther', 5),
        ('Datastore/Postgres/all', 5),
        ('Datastore/Postgres/allOther', 5),
        ('Datastore/operation/Postgres/select', 1),
        ('Datastore/statement/Postgres/%s/select' % DB_SETTINGS['table_name'], 1),
        ('Datastore/operation/Postgres/insert', 1),
        ('Datastore/statement/Postgres/%s/insert' % DB_SETTINGS['table_name'], 1),
        ('Datastore/operation/Postgres/drop', 1),
        ('Datastore/operation/Postgres/create', 1)
)

_disable_scoped_metrics = list(_base_scoped_metrics)
_disable_rollup_metrics = list(_base_rollup_metrics)

_enable_scoped_metrics = list(_base_scoped_metrics)
_enable_rollup_metrics = list(_base_rollup_metrics)

_enable_scoped_metrics.append(('Function/psycopg2:connect', 1))
_disable_scoped_metrics.append(('Function/psycopg2:connect', 1))

_host = instance_hostname(DB_SETTINGS['host'])
_port = DB_SETTINGS['port']

_instance_metric_name = 'Datastore/instance/Postgres/%s/%s' % (_host, _port)

_enable_rollup_metrics.append(
        (_instance_metric_name, 4)
)

_disable_rollup_metrics.append(
        (_instance_metric_name, None)
)

async_keyword_list = ['async', 'async_']


def _exercise_db(async_keyword):
    wait = psycopg2.extras.wait_select

    kwasync = {}
    kwasync[async_keyword] = 1

    async_conn = psycopg2.connect(
            database=DB_SETTINGS['name'], user=DB_SETTINGS['user'],
            password=DB_SETTINGS['password'], host=DB_SETTINGS['host'],
            port=DB_SETTINGS['port'], **kwasync
    )
    wait(async_conn)
    async_cur = async_conn.cursor()

    async_cur.execute("""drop table if exists %s""" % DB_SETTINGS['table_name'])
    wait(async_cur.connection)

    async_cur.execute("""create table %s """ % DB_SETTINGS['table_name'] +
            """(a integer, b real, c text)""")
    wait(async_cur.connection)

    async_cur.execute("""insert into %s """ % DB_SETTINGS['table_name'] +
        """values (%s, %s, %s)""", (1, 1.0, '1.0'))
    wait(async_cur.connection)

    async_cur.execute("""select * from %s""" % DB_SETTINGS['table_name'])
    wait(async_cur.connection)

    for row in async_cur:
        assert isinstance(row, tuple)

    async_conn.close()


# Tests

@override_application_settings(_enable_instance_settings)
@validate_stats_engine_explain_plan_output_is_none()
@validate_transaction_slow_sql_count(num_slow_sql=4)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@validate_transaction_metrics(
        'test_async:test_async_mode_enable_instance',
        scoped_metrics=_enable_scoped_metrics,
        rollup_metrics=_enable_rollup_metrics,
        background_task=True)
@validate_transaction_errors(errors=[])
@background_task()
@pytest.mark.parametrize('async_keyword', async_keyword_list)
def test_async_mode_enable_instance(async_keyword):
    _exercise_db(async_keyword)


@override_application_settings(_disable_instance_settings)
@validate_stats_engine_explain_plan_output_is_none()
@validate_transaction_slow_sql_count(num_slow_sql=4)
@validate_database_trace_inputs(sql_parameters_type=tuple)
@validate_transaction_metrics(
        'test_async:test_async_mode_disable_instance',
        scoped_metrics=_disable_scoped_metrics,
        rollup_metrics=_disable_rollup_metrics,
        background_task=True)
@validate_transaction_errors(errors=[])
@background_task()
@pytest.mark.parametrize('async_keyword', async_keyword_list)
def test_async_mode_disable_instance(async_keyword):
    _exercise_db(async_keyword)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import sys
import time

import pytest
from testing_support.fixtures import capture_transaction_metrics, validate_tt_parenting
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.api.database_trace import database_trace
from newrelic.api.datastore_trace import datastore_trace
from newrelic.api.external_trace import external_trace
from newrelic.api.function_trace import function_trace
from newrelic.api.graphql_trace import graphql_operation_trace, graphql_resolver_trace
from newrelic.api.memcache_trace import memcache_trace
from newrelic.api.message_trace import message_trace

asyncio = pytest.importorskip("asyncio")


@pytest.mark.parametrize(
    "trace,metric",
    [
        (functools.partial(function_trace, name="simple_gen"), "Function/simple_gen"),
        (functools.partial(external_trace, library="lib", url="http://foo.com"), "External/foo.com/lib/"),
        (functools.partial(database_trace, "select * from foo"), "Datastore/statement/None/foo/select"),
        (functools.partial(datastore_trace, "lib", "foo", "bar"), "Datastore/statement/lib/foo/bar"),
        (functools.partial(message_trace, "lib", "op", "typ", "name"), "MessageBroker/lib/typ/op/Named/name"),
        (functools.partial(memcache_trace, "cmd"), "Memcache/cmd"),
        (functools.partial(graphql_operation_trace), "GraphQL/operation/GraphQL/<unknown>/<anonymous>/<unknown>"),
        (functools.partial(graphql_resolver_trace), "GraphQL/resolve/GraphQL/<unknown>"),
    ],
)
def test_async_generator_timing(event_loop, trace, metric):
    @trace()
    async def simple_gen():
        time.sleep(0.1)
        yield
        time.sleep(0.1)

    metrics = []
    full_metrics = {}

    @capture_transaction_metrics(metrics, full_metrics)
    @validate_transaction_metrics(
        "test_async_generator_timing", background_task=True, scoped_metrics=[(metric, 1)], rollup_metrics=[(metric, 1)]
    )
    @background_task(name="test_async_generator_timing")
    def _test_async_generator_timing():
        async def _test():
            async for _ in simple_gen():
                pass

        event_loop.run_until_complete(_test())
    _test_async_generator_timing()

    # Check that coroutines time the total call time (including pauses)
    metric_key = (metric, "")
    assert full_metrics[metric_key].total_call_time >= 0.2


class MyException(Exception):
    pass


@validate_transaction_metrics(
    "test_async_generator_error",
    background_task=True,
    scoped_metrics=[("Function/agen", 1)],
    rollup_metrics=[("Function/agen", 1)],
)
@validate_transaction_errors(errors=["_test_async_generator_trace:MyException"])
def test_async_generator_error(event_loop):
    @function_trace(name="agen")
    async def agen():
        yield

    @background_task(name="test_async_generator_error")
    async def _test():
        gen = agen()
        await gen.asend(None)
        await gen.athrow(MyException)

    with pytest.raises(MyException):
        event_loop.run_until_complete(_test())


@validate_transaction_metrics(
    "test_async_generator_caught_exception",
    background_task=True,
    scoped_metrics=[("Function/agen", 1)],
    rollup_metrics=[("Function/agen", 1)],
)
@validate_transaction_errors(errors=[])
def test_async_generator_caught_exception(event_loop):
    @function_trace(name="agen")
    async def agen():
        for _ in range(2):
            time.sleep(0.1)
            try:
                yield
            except ValueError:
                pass

    metrics = []
    full_metrics = {}

    @capture_transaction_metrics(metrics, full_metrics)
    @background_task(name="test_async_generator_caught_exception")
    def _test_async_generator_caught_exception():
        async def _test():
            gen = agen()
            # kickstart the generator (the try/except logic is inside the
            # generator)
            await gen.asend(None)
            await gen.athrow(ValueError)

            # consume the generator
            async for _ in gen:
                pass

        # The ValueError should not be reraised
        event_loop.run_until_complete(_test())
    _test_async_generator_caught_exception()

    assert full_metrics[("Function/agen", "")].total_call_time >= 0.2


@validate_transaction_metrics(
    "test_async_generator_handles_terminal_nodes",
    background_task=True,
    scoped_metrics=[("Function/parent", 1), ("Function/agen", None)],
    rollup_metrics=[("Function/parent", 1), ("Function/agen", None)],
)
def test_async_generator_handles_terminal_nodes(event_loop):
    # sometimes coroutines can be called underneath terminal nodes
    # In this case, the trace shouldn't actually be created and we also
    # shouldn't get any errors

    @function_trace(name="agen")
    async def agen():
        yield
        time.sleep(0.1)

    @function_trace(name="parent", terminal=True)
    async def parent():
        # parent calls child
        async for _ in agen():
            pass

    metrics = []
    full_metrics = {}

    @capture_transaction_metrics(metrics, full_metrics)
    @background_task(name="test_async_generator_handles_terminal_nodes")
    def _test_async_generator_handles_terminal_nodes():
        async def _test():
            await parent()

        event_loop.run_until_complete(_test())
    _test_async_generator_handles_terminal_nodes()

    metric_key = ("Function/parent", "")
    assert full_metrics[metric_key].total_exclusive_call_time >= 0.1


@validate_transaction_metrics(
    "test_async_generator_close_ends_trace",
    background_task=True,
    scoped_metrics=[("Function/agen", 1)],
    rollup_metrics=[("Function/agen", 1)],
)
def test_async_generator_close_ends_trace(event_loop):
    @function_trace(name="agen")
    async def agen():
        yield

    @background_task(name="test_async_generator_close_ends_trace")
    async def _test():
        gen = agen()

        # kickstart the coroutine
        await gen.asend(None)

        # trace should be ended/recorded by close
        await gen.aclose()

        # We may call gen.close as many times as we want
        await gen.aclose()

    event_loop.run_until_complete(_test())

@validate_tt_parenting(
    (
        "TransactionNode",
        [
            (
                "FunctionNode",
                [
                    ("FunctionNode", []),
                ],
            ),
        ],
    )
)
@validate_transaction_metrics(
    "test_async_generator_parents",
    background_task=True,
    scoped_metrics=[("Function/child", 1), ("Function/parent", 1)],
    rollup_metrics=[("Function/child", 1), ("Function/parent", 1)],
)
def test_async_generator_parents(event_loop):
    @function_trace(name="child")
    async def child():
        yield
        time.sleep(0.1)
        yield

    @function_trace(name="parent")
    async def parent():
        time.sleep(0.1)
        yield
        async for _ in child():
            pass

    metrics = []
    full_metrics = {}

    @capture_transaction_metrics(metrics, full_metrics)
    @background_task(name="test_async_generator_parents")
    def _test_async_generator_parents():
        async def _test():
            async for _ in parent():
                pass

        event_loop.run_until_complete(_test())
    _test_async_generator_parents()

    # Check that the child time is subtracted from the parent time (parenting
    # relationship is correctly established)
    key = ("Function/parent", "")
    assert full_metrics[key].total_exclusive_call_time < 0.2


@validate_transaction_metrics(
    "test_asend_receives_a_value",
    background_task=True,
    scoped_metrics=[("Function/agen", 1)],
    rollup_metrics=[("Function/agen", 1)],
)
def test_asend_receives_a_value(event_loop):
    _received = []
    @function_trace(name="agen")
    async def agen():
        value = yield
        _received.append(value)
        yield value

    @background_task(name="test_asend_receives_a_value")
    async def _test():
        gen = agen()

        # kickstart the coroutine
        await gen.asend(None)

        assert await gen.asend("foobar") == "foobar"
        assert _received and _received[0] == "foobar"

        # finish consumption of the coroutine if necessary
        async for _ in gen:
            pass

    event_loop.run_until_complete(_test())


@validate_transaction_metrics(
    "test_athrow_yields_a_value",
    background_task=True,
    scoped_metrics=[("Function/agen", 1)],
    rollup_metrics=[("Function/agen", 1)],
)
def test_athrow_yields_a_value(event_loop):
    @function_trace(name="agen")
    async def agen():
        for _ in range(2):
            try:
                yield
            except MyException:
                yield "foobar"

    @background_task(name="test_athrow_yields_a_value")
    async def _test():
        gen = agen()

        # kickstart the coroutine
        await gen.asend(None)

        assert await gen.athrow(MyException) == "foobar"

        # finish consumption of the coroutine if necessary
        async for _ in gen:
            pass

    event_loop.run_until_complete(_test())


@validate_transaction_metrics(
    "test_multiple_throws_yield_a_value",
    background_task=True,
    scoped_metrics=[("Function/agen", 1)],
    rollup_metrics=[("Function/agen", 1)],
)
def test_multiple_throws_yield_a_value(event_loop):
    @function_trace(name="agen")
    async def agen():
        value = None
        for _ in range(4):
            try:
                yield value
                value = "bar"
            except MyException:
                value = "foo"


    @background_task(name="test_multiple_throws_yield_a_value")
    async def _test():
        gen = agen()

        # kickstart the coroutine
        assert await gen.asend(None) is None
        assert await gen.athrow(MyException) == "foo"
        assert await gen.athrow(MyException) == "foo"
        assert await gen.asend(None) == "bar"

        # finish consumption of the coroutine if necessary
        async for _ in gen:
            pass

    event_loop.run_until_complete(_test())


@validate_transaction_metrics(
    "test_athrow_does_not_yield_a_value",
    background_task=True,
    scoped_metrics=[("Function/agen", 1)],
    rollup_metrics=[("Function/agen", 1)],
)
def test_athrow_does_not_yield_a_value(event_loop):
    @function_trace(name="agen")
    async def agen():
        for _ in range(2):
            try:
                yield
            except MyException:
                return

    @background_task(name="test_athrow_does_not_yield_a_value")
    async def _test():
        gen = agen()

        # kickstart the coroutine
        await gen.asend(None)

        # async generator will raise StopAsyncIteration
        with pytest.raises(StopAsyncIteration):
            await gen.athrow(MyException)


    event_loop.run_until_complete(_test())


@pytest.mark.parametrize(
    "trace",
    [
        function_trace(name="simple_gen"),
        external_trace(library="lib", url="http://foo.com"),
        database_trace("select * from foo"),
        datastore_trace("lib", "foo", "bar"),
        message_trace("lib", "op", "typ", "name"),
        memcache_trace("cmd"),
    ],
)
def test_async_generator_functions_outside_of_transaction(event_loop, trace):
    @trace
    async def agen():
        for _ in range(2):
            yield "foo"

    async def _test():
        assert [_ async for _ in agen()] == ["foo", "foo"]

    event_loop.run_until_complete(_test())


@validate_transaction_metrics(
    "test_catching_generator_exit_causes_runtime_error",
    background_task=True,
    scoped_metrics=[("Function/agen", 1)],
    rollup_metrics=[("Function/agen", 1)],
)
def test_catching_generator_exit_causes_runtime_error(event_loop):
    @function_trace(name="agen")
    async def agen():
        try:
            yield
        except GeneratorExit:
            yield

    @background_task(name="test_catching_generator_exit_causes_runtime_error")
    async def _test():
        gen = agen()

        # kickstart the coroutine (we're inside the try now)
        await gen.asend(None)

        # Generators cannot catch generator exit exceptions (which are injected by
        # close). This will result in a runtime error.
        with pytest.raises(RuntimeError):
            await gen.aclose()

    event_loop.run_until_complete(_test())


@validate_transaction_metrics(
    "test_async_generator_time_excludes_creation_time",
    background_task=True,
    scoped_metrics=[("Function/agen", 1)],
    rollup_metrics=[("Function/agen", 1)],
)
def test_async_generator_time_excludes_creation_time(event_loop):
    @function_trace(name="agen")
    async def agen():
        yield

    metrics = []
    full_metrics = {}

    @capture_transaction_metrics(metrics, full_metrics)
    @background_task(name="test_async_generator_time_excludes_creation_time")
    def _test_async_generator_time_excludes_creation_time():
        async def _test():
            gen = agen()
            time.sleep(0.1)
            async for _ in gen:
                pass

        event_loop.run_until_complete(_test())
    _test_async_generator_time_excludes_creation_time()

    # check that the trace does not include the time between creation and
    # consumption
    assert full_metrics[("Function/agen", "")].total_call_time < 0.1


@validate_transaction_metrics(
    "test_complete_async_generator",
    background_task=True,
    scoped_metrics=[("Function/agen", 1)],
    rollup_metrics=[("Function/agen", 1)],
)
@background_task(name="test_complete_async_generator")
def test_complete_async_generator(event_loop):
    @function_trace(name="agen")
    async def agen():
        for i in range(5):
            yield i

    async def _test():
        gen = agen()
        assert [x async for x in gen] == [x for x in range(5)]

    event_loop.run_until_complete(_test())


@pytest.mark.parametrize("nr_transaction", [True, False])
def test_incomplete_async_generator(event_loop, nr_transaction):
    @function_trace(name="agen")
    async def agen():
        for _ in range(5):
            yield

    def _test_incomplete_async_generator():
        async def _test():
            c = agen()

            async for _ in c:
                break

        if nr_transaction:
            _test = background_task(name="test_incomplete_async_generator")(_test)

        event_loop.run_until_complete(_test())

    if nr_transaction:
        _test_incomplete_async_generator = validate_transaction_metrics(
            "test_incomplete_async_generator",
            background_task=True,
            scoped_metrics=[("Function/agen", 1)],
            rollup_metrics=[("Function/agen", 1)],
        )(_test_incomplete_async_generator)
    
    _test_incomplete_async_generator()


def test_incomplete_async_generator_transaction_exited(event_loop):
    @function_trace(name="agen")
    async def agen():
        for _ in range(5):
            yield

    @validate_transaction_metrics(
        "test_incomplete_async_generator",
        background_task=True,
        scoped_metrics=[("Function/agen", 1)],
        rollup_metrics=[("Function/agen", 1)],
    )
    def _test_incomplete_async_generator():
        c = agen()
        @background_task(name="test_incomplete_async_generator")
        async def _test():
            async for _ in c:
                break

        event_loop.run_until_complete(_test())

        # Remove generator after transaction completes
        del c

    _test_incomplete_async_generator()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.packages import six

if six.PY3:
    from inspect import Signature

    def bind_args(func, args, kwargs):
        """Bind arguments and apply defaults to missing arguments for a callable."""
        bound_args = Signature.from_callable(func).bind(*args, **kwargs)
        bound_args.apply_defaults()
        return bound_args.arguments

else:
    from inspect import getcallargs

    def bind_args(func, args, kwargs):
        """Bind arguments and apply defaults to missing arguments for a callable."""
        return getcallargs(func, *args, **kwargs)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from testing_support.validators.validate_code_level_metrics import validate_code_level_metrics
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.validators.validate_transaction_errors import validate_transaction_errors

from newrelic.packages import six


@pytest.fixture(autouse=True, scope="module")
def skip_if_not_cornice():
    pytest.importorskip("cornice")


def target_application():
    # We need to delay Pyramid application creation because of ordering
    # issues whereby the agent needs to be initialised before Pyramid is
    # imported and the routes configured. Normally pytest only runs the
    # global fixture which will initialise the agent after each test
    # file is imported, which is too late. We also can't do application
    # creation within a function as Pyramid relies on view handlers being
    # at global scope, so import it from a separate module.

    from _test_application import target_application as _app
    return _app(False, False, True)

_test_cornice_service_scoped_metrics = [
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/pyramid.router:Router.__call__', 1),
        ('Function/_test_application:cornice_service_get_info', 1)]

@validate_code_level_metrics("_test_application", "cornice_service_get_info")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics('_test_application:cornice_service_get_info',
        scoped_metrics=_test_cornice_service_scoped_metrics)
def test_cornice_service():
    application = target_application()
    application.get('/service')

_test_cornice_resource_collection_get_scoped_metrics = [
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/pyramid.router:Router.__call__', 1),
        ('Function/_test_application:Resource.collection_get', 1)]

@validate_code_level_metrics("_test_application.Resource", "collection_get")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics('_test_application:Resource.collection_get',
        scoped_metrics=_test_cornice_resource_collection_get_scoped_metrics)
def test_cornice_resource_collection_get():
    application = target_application()
    application.get('/resource')

_test_cornice_resource_get_scoped_metrics = [
        ('Python/WSGI/Application', 1),
        ('Python/WSGI/Response', 1),
        ('Python/WSGI/Finalize', 1),
        ('Function/pyramid.router:Router.__call__', 1),
        ('Function/_test_application:Resource.get', 1)]

@validate_code_level_metrics("_test_application.Resource", "get")
@validate_transaction_errors(errors=[])
@validate_transaction_metrics('_test_application:Resource.get',
        scoped_metrics=_test_cornice_resource_get_scoped_metrics)
def test_cornice_resource_get():
    application = target_application()
    application.get('/resource/1')

_test_cornice_error_scoped_metrics = [
        ('Python/WSGI/Application', 1),
        ('Function/pyramid.router:Router.__call__', 1),
        ('Function/cornice.pyramidhook:handle_exceptions', 1),
        ('Function/_test_application:cornice_error_get_info', 1)]

if six.PY3:
    _test_cornice_error_errors = ['builtins:RuntimeError']
else:
    _test_cornice_error_errors = ['exceptions:RuntimeError']

@validate_code_level_metrics("_test_application", "cornice_error_get_info")
@validate_transaction_errors(errors=_test_cornice_error_errors)
@validate_transaction_metrics('_test_application:cornice_error_get_info',
        scoped_metrics=_test_cornice_error_scoped_metrics)
def test_cornice_error():
    application = target_application()
    with pytest.raises(RuntimeError):
        application.get('/cornice_error', status=500)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import django
from testing_support.fixtures import (
    override_application_settings,
    override_generic_settings,
    override_ignore_status_codes,
)
from testing_support.validators.validate_code_level_metrics import (
    validate_code_level_metrics,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.hooks.framework_django import django_settings

DJANGO_VERSION = tuple(map(int, django.get_version().split(".")[:2]))
DJANGO_SETTINGS_MODULE = os.environ.get("DJANGO_SETTINGS_MODULE", None)


def target_application():
    from _target_application import _target_application

    return _target_application


# The middleware scoped metrics are dependent on the MIDDLEWARE_CLASSES or
# MIDDLEWARE defined in the version-specific Django settings.py file.

_test_django_pre_1_10_middleware_scoped_metrics = [
    (("Function/django.middleware.common:" "CommonMiddleware.process_request"), 1),
    (("Function/django.contrib.sessions.middleware:" "SessionMiddleware.process_request"), 1),
    (("Function/django.contrib.auth.middleware:" "AuthenticationMiddleware.process_request"), 1),
    (("Function/django.contrib.messages.middleware:" "MessageMiddleware.process_request"), 1),
    (("Function/django.middleware.csrf:" "CsrfViewMiddleware.process_view"), 1),
    (("Function/django.contrib.messages.middleware:" "MessageMiddleware.process_response"), 1),
    (("Function/django.middleware.csrf:" "CsrfViewMiddleware.process_response"), 1),
    (("Function/django.contrib.sessions.middleware:" "SessionMiddleware.process_response"), 1),
    (("Function/django.middleware.common:" "CommonMiddleware.process_response"), 1),
    (("Function/django.middleware.gzip:" "GZipMiddleware.process_response"), 1),
    (("Function/newrelic.hooks.framework_django:" "browser_timing_insertion"), 1),
]

_test_django_post_1_10_middleware_scoped_metrics = [
    ("Function/django.middleware.security:SecurityMiddleware", 1),
    ("Function/django.contrib.sessions.middleware:SessionMiddleware", 1),
    ("Function/django.middleware.common:CommonMiddleware", 1),
    ("Function/django.middleware.csrf:CsrfViewMiddleware", 1),
    ("Function/django.contrib.auth.middleware:AuthenticationMiddleware", 1),
    ("Function/django.contrib.messages.middleware:MessageMiddleware", 1),
    ("Function/django.middleware.clickjacking:XFrameOptionsMiddleware", 1),
    ("Function/django.middleware.gzip:GZipMiddleware", 1),
]

_test_django_pre_1_10_url_resolver_scoped_metrics = [
    ("Function/django.core.urlresolvers:RegexURLResolver.resolve", "present"),
]

_test_django_post_1_10_url_resolver_scoped_metrics = [
    ("Function/django.urls.resolvers:RegexURLResolver.resolve", "present"),
]

_test_django_post_2_0_url_resolver_scoped_metrics = [
    ("Function/django.urls.resolvers:URLResolver.resolve", "present"),
]

_test_application_index_scoped_metrics = [
    ("Function/django.core.handlers.wsgi:WSGIHandler.__call__", 1),
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/views:index", 1),
]

if DJANGO_VERSION >= (1, 5):
    _test_application_index_scoped_metrics.extend([("Function/django.http.response:HttpResponse.close", 1)])

if DJANGO_VERSION < (1, 10):
    _test_application_index_scoped_metrics.extend(_test_django_pre_1_10_url_resolver_scoped_metrics)
elif DJANGO_VERSION >= (2, 0):
    _test_application_index_scoped_metrics.extend(_test_django_post_2_0_url_resolver_scoped_metrics)
else:
    _test_application_index_scoped_metrics.extend(_test_django_post_1_10_url_resolver_scoped_metrics)

if DJANGO_SETTINGS_MODULE == "settings_0110_old":
    _test_application_index_scoped_metrics.extend(_test_django_pre_1_10_middleware_scoped_metrics)
elif DJANGO_SETTINGS_MODULE == "settings_0110_new":
    _test_application_index_scoped_metrics.extend(_test_django_post_1_10_middleware_scoped_metrics)
elif DJANGO_VERSION < (1, 10):
    _test_application_index_scoped_metrics.extend(_test_django_pre_1_10_middleware_scoped_metrics)


@validate_transaction_errors(errors=[])
@validate_transaction_metrics("views:index", scoped_metrics=_test_application_index_scoped_metrics)
@validate_code_level_metrics("views", "index")
def test_application_index():
    test_application = target_application()
    response = test_application.get("")
    response.mustcontain("INDEX RESPONSE")


@validate_transaction_metrics("views:exception")
@validate_code_level_metrics("views", "exception")
def test_application_exception():
    test_application = target_application()
    test_application.get("/exception", status=500)


_test_application_not_found_scoped_metrics = [
    ("Function/django.core.handlers.wsgi:WSGIHandler.__call__", 1),
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
]

if DJANGO_VERSION >= (1, 5):
    _test_application_not_found_scoped_metrics.extend([("Function/django.http.response:HttpResponseNotFound.close", 1)])

if DJANGO_VERSION < (1, 10):
    _test_application_not_found_scoped_metrics.extend(_test_django_pre_1_10_url_resolver_scoped_metrics)
elif DJANGO_VERSION >= (2, 0):
    _test_application_not_found_scoped_metrics.extend(_test_django_post_2_0_url_resolver_scoped_metrics)
else:
    _test_application_not_found_scoped_metrics.extend(_test_django_post_1_10_url_resolver_scoped_metrics)

if DJANGO_SETTINGS_MODULE == "settings_0110_old":
    _test_application_not_found_scoped_metrics.extend(_test_django_pre_1_10_middleware_scoped_metrics)
    # The `CsrfViewMiddleware.process_view` isn't called for 404 Not Found.
    _test_application_not_found_scoped_metrics.remove(
        ("Function/django.middleware.csrf:CsrfViewMiddleware.process_view", 1)
    )
elif DJANGO_SETTINGS_MODULE == "settings_0110_new":
    _test_application_not_found_scoped_metrics.extend(_test_django_post_1_10_middleware_scoped_metrics)
elif DJANGO_VERSION < (1, 10):
    _test_application_not_found_scoped_metrics.extend(_test_django_pre_1_10_middleware_scoped_metrics)
    # The `CsrfViewMiddleware.process_view` isn't called for 404 Not Found.
    _test_application_not_found_scoped_metrics.remove(
        ("Function/django.middleware.csrf:CsrfViewMiddleware.process_view", 1)
    )


@validate_transaction_errors(errors=[])
@validate_transaction_metrics(
    "django.views.debug:technical_404_response", scoped_metrics=_test_application_not_found_scoped_metrics
)
def test_application_not_found():
    test_application = target_application()
    test_application.get("/not_found", status=404)


@override_ignore_status_codes([403])
@validate_transaction_errors(errors=[])
@validate_transaction_metrics("views:permission_denied")
@validate_code_level_metrics("views", "permission_denied")
def test_ignored_status_code():
    test_application = target_application()
    test_application.get("/permission_denied", status=403)


@override_ignore_status_codes([410])
@validate_transaction_errors(errors=[])
@validate_transaction_metrics("views:middleware_410")
@validate_code_level_metrics("views", "middleware_410")
def test_middleware_ignore_status_codes():
    test_application = target_application()
    test_application.get("/middleware_410", status=410)


_test_application_cbv_scoped_metrics = [
    ("Function/django.core.handlers.wsgi:WSGIHandler.__call__", 1),
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/views:MyView", 1),
    ("Function/views:MyView.get", 1),
]

if DJANGO_VERSION >= (1, 5):
    _test_application_cbv_scoped_metrics.extend([("Function/django.http.response:HttpResponse.close", 1)])

if DJANGO_VERSION < (1, 10):
    _test_application_cbv_scoped_metrics.extend(_test_django_pre_1_10_url_resolver_scoped_metrics)
elif DJANGO_VERSION >= (2, 0):
    _test_application_cbv_scoped_metrics.extend(_test_django_post_2_0_url_resolver_scoped_metrics)
else:
    _test_application_cbv_scoped_metrics.extend(_test_django_post_1_10_url_resolver_scoped_metrics)

if DJANGO_SETTINGS_MODULE == "settings_0110_old":
    _test_application_cbv_scoped_metrics.extend(_test_django_pre_1_10_middleware_scoped_metrics)
elif DJANGO_SETTINGS_MODULE == "settings_0110_new":
    _test_application_cbv_scoped_metrics.extend(_test_django_post_1_10_middleware_scoped_metrics)
elif DJANGO_VERSION < (1, 10):
    _test_application_cbv_scoped_metrics.extend(_test_django_pre_1_10_middleware_scoped_metrics)


@validate_transaction_errors(errors=[])
@validate_transaction_metrics("views:MyView.get", scoped_metrics=_test_application_cbv_scoped_metrics)
@validate_code_level_metrics("views.MyView", "get")
def test_application_cbv():
    test_application = target_application()
    response = test_application.get("/cbv")
    response.mustcontain("CBV RESPONSE")


_test_application_deferred_cbv_scoped_metrics = [
    ("Function/django.core.handlers.wsgi:WSGIHandler.__call__", 1),
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/views:deferred_cbv", 1),
    ("Function/views:MyView.get", 1),
]

if DJANGO_VERSION >= (1, 5):
    _test_application_deferred_cbv_scoped_metrics.extend([("Function/django.http.response:HttpResponse.close", 1)])

if DJANGO_VERSION < (1, 10):
    _test_application_deferred_cbv_scoped_metrics.extend(_test_django_pre_1_10_url_resolver_scoped_metrics)
elif DJANGO_VERSION >= (2, 0):
    _test_application_deferred_cbv_scoped_metrics.extend(_test_django_post_2_0_url_resolver_scoped_metrics)
else:
    _test_application_deferred_cbv_scoped_metrics.extend(_test_django_post_1_10_url_resolver_scoped_metrics)

if DJANGO_SETTINGS_MODULE == "settings_0110_old":
    _test_application_deferred_cbv_scoped_metrics.extend(_test_django_pre_1_10_middleware_scoped_metrics)
elif DJANGO_SETTINGS_MODULE == "settings_0110_new":
    _test_application_deferred_cbv_scoped_metrics.extend(_test_django_post_1_10_middleware_scoped_metrics)
elif DJANGO_VERSION < (1, 10):
    _test_application_deferred_cbv_scoped_metrics.extend(_test_django_pre_1_10_middleware_scoped_metrics)


@validate_transaction_errors(errors=[])
@validate_transaction_metrics("views:deferred_cbv", scoped_metrics=_test_application_deferred_cbv_scoped_metrics)
@validate_code_level_metrics("views", "deferred_cbv")
def test_application_deferred_cbv():
    test_application = target_application()
    response = test_application.get("/deferred_cbv")
    response.mustcontain("CBV RESPONSE")


_test_html_insertion_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion_django_middleware():
    test_application = target_application()
    response = test_application.get("/html_insertion", status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain("NREUM HEADER", "NREUM.info")


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion_django_gzip_middleware_enabled():
    test_application = target_application()

    # GZipMiddleware only fires if given the following header.

    gzip_header = {"Accept-Encoding": "gzip"}
    response = test_application.get("/gzip_html_insertion", status=200, headers=gzip_header)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    # The response.text will already be gunzipped

    response.mustcontain("NREUM HEADER", "NREUM.info")


_test_html_insertion_settings_disabled = {
    "browser_monitoring.enabled": False,
    "browser_monitoring.auto_instrument": False,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_settings_disabled)
def test_html_insertion_django_gzip_middleware_disabled():
    test_application = target_application()

    # GZipMiddleware only fires if given the following header.

    gzip_header = {"Accept-Encoding": "gzip"}
    response = test_application.get("/gzip_html_insertion", status=200, headers=gzip_header)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    # The response.text will already be gunzipped

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])


_test_html_insertion_manual_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": True,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_manual_settings)
def test_html_insertion_manual_django_middleware():
    test_application = target_application()
    response = test_application.get("/html_insertion_manual", status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion_unnamed_attachment_header_django_middleware():
    test_application = target_application()
    response = test_application.get("/html_insertion_unnamed_attachment_header", status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion_named_attachment_header_django_middleware():
    test_application = target_application()
    response = test_application.get("/html_insertion_named_attachment_header", status=200)

    # The 'NREUM HEADER' value comes from our override for the header.
    # The 'NREUM.info' value comes from the programmatically generated
    # header added by the agent.

    response.mustcontain(no=["NREUM HEADER", "NREUM.info"])


_test_html_insertion_settings = {
    "browser_monitoring.enabled": True,
    "browser_monitoring.auto_instrument": False,
    "js_agent_loader": "<!-- NREUM HEADER -->",
}


@override_application_settings(_test_html_insertion_settings)
def test_html_insertion_manual_tag_instrumentation():
    test_application = target_application()
    response = test_application.get("/template_tags")

    # Assert that the instrumentation is not inappropriately escaped

    response.mustcontain("<!-- NREUM HEADER -->", no=["&lt;!-- NREUM HEADER --&gt"])


_test_application_inclusion_tag_scoped_metrics = [
    ("Function/django.core.handlers.wsgi:WSGIHandler.__call__", 1),
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/views:inclusion_tag", 1),
    ("Template/Render/main.html", 1),
]

if DJANGO_VERSION < (1, 9):
    _test_application_inclusion_tag_scoped_metrics.extend([("Template/Include/results.html", 1)])

if DJANGO_VERSION < (1, 10):
    _test_application_inclusion_tag_scoped_metrics.extend(_test_django_pre_1_10_url_resolver_scoped_metrics)
elif DJANGO_VERSION >= (2, 0):
    _test_application_inclusion_tag_scoped_metrics.extend(_test_django_post_2_0_url_resolver_scoped_metrics)
else:
    _test_application_inclusion_tag_scoped_metrics.extend(_test_django_post_1_10_url_resolver_scoped_metrics)

if DJANGO_SETTINGS_MODULE == "settings_0110_old":
    _test_application_inclusion_tag_scoped_metrics.extend(_test_django_pre_1_10_middleware_scoped_metrics)
elif DJANGO_SETTINGS_MODULE == "settings_0110_new":
    _test_application_inclusion_tag_scoped_metrics.extend(_test_django_post_1_10_middleware_scoped_metrics)
elif DJANGO_VERSION < (1, 10):
    _test_application_inclusion_tag_scoped_metrics.extend(_test_django_pre_1_10_middleware_scoped_metrics)

try:
    _test_application_inclusion_tag_scoped_metrics.remove(
        (("Function/newrelic.hooks.framework_django:" "browser_timing_insertion"), 1)
    )
except ValueError:
    pass


@validate_transaction_errors(errors=[])
@validate_transaction_metrics("views:inclusion_tag", scoped_metrics=_test_application_inclusion_tag_scoped_metrics)
@validate_code_level_metrics("views", "inclusion_tag")
def test_application_inclusion_tag():
    test_application = target_application()
    response = test_application.get("/inclusion_tag")
    response.mustcontain("Inclusion tag")


_test_inclusion_tag_template_tags_scoped_metrics = [
    ("Function/django.core.handlers.wsgi:WSGIHandler.__call__", 1),
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    ("Function/views:inclusion_tag", 1),
    ("Template/Render/main.html", 1),
]

if DJANGO_VERSION < (1, 9):
    _test_inclusion_tag_template_tags_scoped_metrics.extend(
        [("Template/Include/results.html", 1), ("Template/Tag/show_results", 1)]
    )

_test_inclusion_tag_settings = {"instrumentation.templates.inclusion_tag": "*"}

if DJANGO_VERSION < (1, 10):
    _test_inclusion_tag_template_tags_scoped_metrics.extend(_test_django_pre_1_10_url_resolver_scoped_metrics)
elif DJANGO_VERSION >= (2, 0):
    _test_inclusion_tag_template_tags_scoped_metrics.extend(_test_django_post_2_0_url_resolver_scoped_metrics)
else:
    _test_inclusion_tag_template_tags_scoped_metrics.extend(_test_django_post_1_10_url_resolver_scoped_metrics)

if DJANGO_SETTINGS_MODULE == "settings_0110_old":
    _test_inclusion_tag_template_tags_scoped_metrics.extend(_test_django_pre_1_10_middleware_scoped_metrics)
elif DJANGO_SETTINGS_MODULE == "settings_0110_new":
    _test_inclusion_tag_template_tags_scoped_metrics.extend(_test_django_post_1_10_middleware_scoped_metrics)
elif DJANGO_VERSION < (1, 10):
    _test_inclusion_tag_template_tags_scoped_metrics.extend(_test_django_pre_1_10_middleware_scoped_metrics)

try:
    _test_inclusion_tag_template_tags_scoped_metrics.remove(
        (("Function/newrelic.hooks.framework_django:" "browser_timing_insertion"), 1)
    )
except ValueError:
    pass


@validate_transaction_errors(errors=[])
@validate_transaction_metrics("views:inclusion_tag", scoped_metrics=_test_inclusion_tag_template_tags_scoped_metrics)
@override_generic_settings(django_settings, _test_inclusion_tag_settings)
@validate_code_level_metrics("views", "inclusion_tag")
def test_inclusion_tag_template_tag_metric():
    test_application = target_application()
    response = test_application.get("/inclusion_tag")
    response.mustcontain("Inclusion tag")


_test_template_render_exception_scoped_metrics_base = [
    ("Function/django.core.handlers.wsgi:WSGIHandler.__call__", 1),
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
]

if DJANGO_VERSION < (1, 5):
    _test_template_render_exception_scoped_metrics_base.append(
        ("Function/django.http:HttpResponseServerError.close", 1)
    )
elif DJANGO_VERSION < (1, 8):
    _test_template_render_exception_scoped_metrics_base.append(
        ("Function/django.http.response:HttpResponseServerError.close", 1)
    )
else:
    _test_template_render_exception_scoped_metrics_base.append(("Function/django.http.response:HttpResponse.close", 1))

if DJANGO_VERSION < (1, 10):
    _test_template_render_exception_scoped_metrics_base.extend(_test_django_pre_1_10_url_resolver_scoped_metrics)
elif DJANGO_VERSION >= (2, 0):
    _test_template_render_exception_scoped_metrics_base.extend(_test_django_post_2_0_url_resolver_scoped_metrics)
else:
    _test_template_render_exception_scoped_metrics_base.extend(_test_django_post_1_10_url_resolver_scoped_metrics)

if DJANGO_SETTINGS_MODULE == "settings_0110_old":
    _test_template_render_exception_scoped_metrics_base.extend(_test_django_pre_1_10_middleware_scoped_metrics)
elif DJANGO_SETTINGS_MODULE == "settings_0110_new":
    _test_template_render_exception_scoped_metrics_base.extend(_test_django_post_1_10_middleware_scoped_metrics)
elif DJANGO_VERSION < (1, 10):
    _test_template_render_exception_scoped_metrics_base.extend(_test_django_pre_1_10_middleware_scoped_metrics)

if DJANGO_VERSION < (1, 9):
    _test_template_render_exception_errors = ["django.template.base:TemplateSyntaxError"]
else:
    _test_template_render_exception_errors = ["django.template.exceptions:TemplateSyntaxError"]

_test_template_render_exception_function_scoped_metrics = list(_test_template_render_exception_scoped_metrics_base)
_test_template_render_exception_function_scoped_metrics.extend(
    [
        ("Function/views:render_exception_function", 1),
    ]
)


@validate_transaction_errors(errors=_test_template_render_exception_errors)
@validate_transaction_metrics(
    "views:render_exception_function", scoped_metrics=_test_template_render_exception_function_scoped_metrics
)
@validate_code_level_metrics("views", "render_exception_function")
def test_template_render_exception_function():
    test_application = target_application()
    test_application.get("/render_exception_function", status=500)


_test_template_render_exception_class_scoped_metrics = list(_test_template_render_exception_scoped_metrics_base)
_test_template_render_exception_class_scoped_metrics.extend(
    [
        ("Function/views:RenderExceptionClass", 1),
        ("Function/views:RenderExceptionClass.get", 1),
    ]
)


@validate_transaction_errors(errors=_test_template_render_exception_errors)
@validate_transaction_metrics(
    "views:RenderExceptionClass.get", scoped_metrics=_test_template_render_exception_class_scoped_metrics
)
@validate_code_level_metrics("views.RenderExceptionClass", "get")
def test_template_render_exception_class():
    test_application = target_application()
    test_application.get("/render_exception_class", status=500)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

INI_FILE_EMPTY = b"""
[newrelic]
"""


INI_FILE_W3C = b"""
[newrelic]
distributed_tracing.exclude_newrelic_header = true
"""

# Tests for loading settings and testing for values precedence
@pytest.mark.parametrize('ini,env,expected_format', (
    (INI_FILE_EMPTY, {}, False),
    (INI_FILE_W3C, {}, True),
))
def test_distributed_trace_setings(ini, env, expected_format, global_settings):

    settings = global_settings()
    assert settings.distributed_tracing.exclude_newrelic_header == expected_format



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from sklearn import __version__  # noqa: this is needed for get_package_version
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.common.package_version_utils import get_package_version
from newrelic.packages import six

SKLEARN_VERSION = tuple(map(int, get_package_version("sklearn").split(".")))


@pytest.mark.parametrize(
    "cluster_model_name",
    [
        "AffinityPropagation",
        "AgglomerativeClustering",
        "Birch",
        "DBSCAN",
        "FeatureAgglomeration",
        "KMeans",
        "MeanShift",
        "MiniBatchKMeans",
        "SpectralBiclustering",
        "SpectralCoclustering",
        "SpectralClustering",
    ],
)
def test_below_v1_1_model_methods_wrapped_in_function_trace(cluster_model_name, run_cluster_model):
    expected_scoped_metrics = {
        "AffinityPropagation": [
            ("Function/MLModel/Sklearn/Named/AffinityPropagation.fit", 2),
            ("Function/MLModel/Sklearn/Named/AffinityPropagation.predict", 1),
            ("Function/MLModel/Sklearn/Named/AffinityPropagation.fit_predict", 1),
        ],
        "AgglomerativeClustering": [
            ("Function/MLModel/Sklearn/Named/AgglomerativeClustering.fit", 2),
            ("Function/MLModel/Sklearn/Named/AgglomerativeClustering.fit_predict", 1),
        ],
        "Birch": [
            ("Function/MLModel/Sklearn/Named/Birch.fit", 2),
            (
                "Function/MLModel/Sklearn/Named/Birch.predict",
                1 if SKLEARN_VERSION >= (1, 0, 0) else 3,
            ),
            ("Function/MLModel/Sklearn/Named/Birch.fit_predict", 1),
            ("Function/MLModel/Sklearn/Named/Birch.transform", 1),
        ],
        "DBSCAN": [
            ("Function/MLModel/Sklearn/Named/DBSCAN.fit", 2),
            ("Function/MLModel/Sklearn/Named/DBSCAN.fit_predict", 1),
        ],
        "FeatureAgglomeration": [
            ("Function/MLModel/Sklearn/Named/FeatureAgglomeration.fit", 1),
            ("Function/MLModel/Sklearn/Named/FeatureAgglomeration.transform", 1),
        ],
        "KMeans": [
            ("Function/MLModel/Sklearn/Named/KMeans.fit", 2),
            ("Function/MLModel/Sklearn/Named/KMeans.predict", 1),
            ("Function/MLModel/Sklearn/Named/KMeans.fit_predict", 1),
            ("Function/MLModel/Sklearn/Named/KMeans.transform", 1),
        ],
        "MeanShift": [
            ("Function/MLModel/Sklearn/Named/MeanShift.fit", 2),
            ("Function/MLModel/Sklearn/Named/MeanShift.predict", 1),
            ("Function/MLModel/Sklearn/Named/MeanShift.fit_predict", 1),
        ],
        "MiniBatchKMeans": [
            ("Function/MLModel/Sklearn/Named/MiniBatchKMeans.fit", 2),
            ("Function/MLModel/Sklearn/Named/MiniBatchKMeans.predict", 1),
            ("Function/MLModel/Sklearn/Named/MiniBatchKMeans.fit_predict", 1),
        ],
        "SpectralBiclustering": [
            ("Function/MLModel/Sklearn/Named/SpectralBiclustering.fit", 1),
        ],
        "SpectralCoclustering": [
            ("Function/MLModel/Sklearn/Named/SpectralCoclustering.fit", 1),
        ],
        "SpectralClustering": [
            ("Function/MLModel/Sklearn/Named/SpectralClustering.fit", 2),
            ("Function/MLModel/Sklearn/Named/SpectralClustering.fit_predict", 1),
        ],
    }
    expected_transaction_name = "test_cluster_models:_test"
    if six.PY3:
        expected_transaction_name = (
            "test_cluster_models:test_below_v1_1_model_methods_wrapped_in_function_trace.<locals>._test"
        )

    @validate_transaction_metrics(
        expected_transaction_name,
        scoped_metrics=expected_scoped_metrics[cluster_model_name],
        rollup_metrics=expected_scoped_metrics[cluster_model_name],
        background_task=True,
    )
    @background_task()
    def _test():
        run_cluster_model(cluster_model_name)

    _test()


@pytest.mark.skipif(SKLEARN_VERSION < (1, 1, 0), reason="Requires sklearn > 1.1")
@pytest.mark.parametrize(
    "cluster_model_name",
    [
        "BisectingKMeans",
        "OPTICS",
    ],
)
def test_above_v1_1_model_methods_wrapped_in_function_trace(cluster_model_name, run_cluster_model):
    expected_scoped_metrics = {
        "BisectingKMeans": [
            ("Function/MLModel/Sklearn/Named/BisectingKMeans.fit", 2),
            ("Function/MLModel/Sklearn/Named/BisectingKMeans.predict", 1),
            ("Function/MLModel/Sklearn/Named/BisectingKMeans.fit_predict", 1),
        ],
        "OPTICS": [
            ("Function/MLModel/Sklearn/Named/OPTICS.fit", 2),
            ("Function/MLModel/Sklearn/Named/OPTICS.fit_predict", 1),
        ],
    }
    expected_transaction_name = "test_cluster_models:_test"
    if six.PY3:
        expected_transaction_name = (
            "test_cluster_models:test_above_v1_1_model_methods_wrapped_in_function_trace.<locals>._test"
        )

    @validate_transaction_metrics(
        expected_transaction_name,
        scoped_metrics=expected_scoped_metrics[cluster_model_name],
        rollup_metrics=expected_scoped_metrics[cluster_model_name],
        background_task=True,
    )
    @background_task()
    def _test():
        run_cluster_model(cluster_model_name)

    _test()


@pytest.fixture
def run_cluster_model():
    def _run(cluster_model_name):
        import sklearn.cluster
        from sklearn.datasets import load_iris
        from sklearn.model_selection import train_test_split

        X, y = load_iris(return_X_y=True)
        x_train, x_test, y_train, y_test = train_test_split(X, y, stratify=y, random_state=0)

        clf = getattr(sklearn.cluster, cluster_model_name)()

        model = clf.fit(x_train, y_train)

        if hasattr(model, "predict"):
            model.predict(x_test)
        if hasattr(model, "score"):
            model.score(x_test, y_test)
        if hasattr(model, "fit_predict"):
            model.fit_predict(x_test)
        if hasattr(model, "predict_log_proba"):
            model.predict_log_proba(x_test)
        if hasattr(model, "predict_proba"):
            model.predict_proba(x_test)
        if hasattr(model, "transform"):
            model.transform(x_test)

        return model

    return _run



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from graphql.language.source import Source

from ._target_schema_async import target_schema as target_schema_async
from ._target_schema_sync import target_schema as target_schema_sync


def check_response(query, response):
    if isinstance(query, str) and "error" not in query or isinstance(query, Source) and "error" not in query.body:
        assert not response.errors, response.errors
        assert response.data
    else:
        assert response.errors


def run_sync(schema):
    def _run_sync(query, middleware=None):
        try:
            from graphql import graphql_sync as graphql
        except ImportError:
            from graphql import graphql

        response = graphql(schema, query, middleware=middleware)

        check_response(query, response)

        return response.data

    return _run_sync


def run_async(schema):
    import asyncio

    from graphql import graphql

    def _run_async(query, middleware=None):
        coro = graphql(schema, query, middleware=middleware)
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(coro)

        check_response(query, response)

        return response.data

    return _run_async


target_application = {
    "sync-sync": run_sync(target_schema_sync),
    "async-sync": run_async(target_schema_sync),
    "async-async": run_async(target_schema_async),
}



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import uuid

import botocore.session
import moto
from testing_support.fixtures import dt_enabled
from testing_support.validators.validate_span_events import validate_span_events
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)
from testing_support.validators.validate_tt_segment_params import (
    validate_tt_segment_params,
)

from newrelic.api.background_task import background_task
from newrelic.common.package_version_utils import get_package_version_tuple

MOTO_VERSION = get_package_version_tuple("moto")

# patch earlier versions of moto to support py37
if sys.version_info >= (3, 7) and MOTO_VERSION <= (1, 3, 1):
    import re

    moto.packages.responses.responses.re._pattern_type = re.Pattern


AWS_ACCESS_KEY_ID = "AAAAAAAAAAAACCESSKEY"
AWS_SECRET_ACCESS_KEY = "AAAAAASECRETKEY"  # nosec (This is fine for testing purposes)
AWS_REGION = "us-east-1"

TEST_TABLE = "python-agent-test-%s" % uuid.uuid4()


_dynamodb_scoped_metrics = [
    ("Datastore/statement/DynamoDB/%s/create_table" % TEST_TABLE, 1),
    ("Datastore/statement/DynamoDB/%s/put_item" % TEST_TABLE, 1),
    ("Datastore/statement/DynamoDB/%s/get_item" % TEST_TABLE, 1),
    ("Datastore/statement/DynamoDB/%s/update_item" % TEST_TABLE, 1),
    ("Datastore/statement/DynamoDB/%s/query" % TEST_TABLE, 1),
    ("Datastore/statement/DynamoDB/%s/scan" % TEST_TABLE, 1),
    ("Datastore/statement/DynamoDB/%s/delete_item" % TEST_TABLE, 1),
    ("Datastore/statement/DynamoDB/%s/delete_table" % TEST_TABLE, 1),
]

_dynamodb_rollup_metrics = [
    ("Datastore/all", 8),
    ("Datastore/allOther", 8),
    ("Datastore/DynamoDB/all", 8),
    ("Datastore/DynamoDB/allOther", 8),
]


@dt_enabled
@validate_span_events(expected_agents=("aws.requestId",), count=8)
@validate_span_events(exact_agents={"aws.operation": "PutItem"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "GetItem"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "DeleteItem"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "CreateTable"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "DeleteTable"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "Query"}, count=1)
@validate_span_events(exact_agents={"aws.operation": "Scan"}, count=1)
@validate_tt_segment_params(present_params=("aws.requestId",))
@validate_transaction_metrics(
    "test_botocore_dynamodb:test_dynamodb",
    scoped_metrics=_dynamodb_scoped_metrics,
    rollup_metrics=_dynamodb_rollup_metrics,
    background_task=True,
)
@background_task()
@moto.mock_dynamodb
def test_dynamodb():
    session = botocore.session.get_session()
    client = session.create_client(
        "dynamodb",
        region_name=AWS_REGION,
        aws_access_key_id=AWS_ACCESS_KEY_ID,
        aws_secret_access_key=AWS_SECRET_ACCESS_KEY,
    )

    # Create table
    resp = client.create_table(
        TableName=TEST_TABLE,
        AttributeDefinitions=[
            {"AttributeName": "Id", "AttributeType": "N"},
            {"AttributeName": "Foo", "AttributeType": "S"},
        ],
        KeySchema=[
            {"AttributeName": "Id", "KeyType": "HASH"},
            {"AttributeName": "Foo", "KeyType": "RANGE"},
        ],
        ProvisionedThroughput={
            "ReadCapacityUnits": 5,
            "WriteCapacityUnits": 5,
        },
    )
    assert resp["TableDescription"]["TableName"] == TEST_TABLE
    # moto response is ACTIVE, AWS response is CREATING
    # assert resp['TableDescription']['TableStatus'] == 'ACTIVE'

    # # AWS needs time to create the table
    # import time
    # time.sleep(15)

    # Put item
    resp = client.put_item(
        TableName=TEST_TABLE,
        Item={
            "Id": {"N": "101"},
            "Foo": {"S": "hello_world"},
            "SomeValue": {"S": "some_random_attribute"},
        },
    )
    # No checking response, due to inconsistent return values.
    # moto returns resp['Attributes']. AWS returns resp['ResponseMetadata']

    # Get item
    resp = client.get_item(
        TableName=TEST_TABLE,
        Key={
            "Id": {"N": "101"},
            "Foo": {"S": "hello_world"},
            "SomeValue": {"S": "some_random_attribute"},
        },
    )
    assert resp["Item"]["SomeValue"]["S"] == "some_random_attribute"

    # Update item
    resp = client.update_item(
        TableName=TEST_TABLE,
        Key={
            "Id": {"N": "101"},
            "Foo": {"S": "hello_world"},
            "SomeValue": {"S": "some_random_attribute"},
        },
        AttributeUpdates={
            "Foo2": {"Value": {"S": "hello_world2"}, "Action": "PUT"},
        },
        ReturnValues="ALL_NEW",
    )
    assert resp["Attributes"]["Foo2"]

    # Query for item
    resp = client.query(
        TableName=TEST_TABLE,
        Select="ALL_ATTRIBUTES",
        KeyConditionExpression="#Id = :v_id",
        ExpressionAttributeNames={"#Id": "Id"},
        ExpressionAttributeValues={":v_id": {"N": "101"}},
    )
    assert len(resp["Items"]) == 1
    assert resp["Items"][0]["SomeValue"]["S"] == "some_random_attribute"

    # Scan
    resp = client.scan(TableName=TEST_TABLE)
    assert len(resp["Items"]) == 1

    # Delete item
    resp = client.delete_item(
        TableName=TEST_TABLE,
        Key={
            "Id": {"N": "101"},
            "Foo": {"S": "hello_world"},
        },
    )
    # No checking response, due to inconsistent return values.
    # moto returns resp['Attributes']. AWS returns resp['ResponseMetadata']

    # Delete table
    resp = client.delete_table(TableName=TEST_TABLE)
    assert resp["TableDescription"]["TableName"] == TEST_TABLE
    # moto response is ACTIVE, AWS response is DELETING
    # assert resp['TableDescription']['TableStatus'] == 'DELETING'



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from testing_support.fixtures import reset_core_stats_engine
from testing_support.validators.validate_custom_metrics_outside_transaction import (
    validate_custom_metrics_outside_transaction,
)

from newrelic.api.application import application_instance as application
from newrelic.api.background_task import background_task
from newrelic.api.transaction import (
    current_transaction,
    record_custom_metric,
    record_custom_metrics,
)


# Testing record_custom_metric
@reset_core_stats_engine()
@background_task()
def test_custom_metric_inside_transaction():
    transaction = current_transaction()
    record_custom_metric("CustomMetric/InsideTransaction/Count", 1)
    for metric in transaction._custom_metrics.metrics():
        assert metric == ("CustomMetric/InsideTransaction/Count", [1, 1, 1, 1, 1, 1])


@reset_core_stats_engine()
@validate_custom_metrics_outside_transaction([("CustomMetric/OutsideTransaction/Count", 1)])
@background_task()
def test_custom_metric_outside_transaction_with_app():
    app = application()
    record_custom_metric("CustomMetric/OutsideTransaction/Count", 1, application=app)


# Testing record_custom_metricS
@reset_core_stats_engine()
@background_task()
def test_custom_metrics_inside_transaction():
    transaction = current_transaction()
    record_custom_metrics([("CustomMetrics/InsideTransaction/Count", 1)])
    for metric in transaction._custom_metrics.metrics():
        assert metric == ("CustomMetrics/InsideTransaction/Count", [1, 1, 1, 1, 1, 1])


@reset_core_stats_engine()
@validate_custom_metrics_outside_transaction([("CustomMetrics/OutsideTransaction/Count", 1)])
@background_task()
def test_custom_metrics_outside_transaction_with_app():
    app = application()
    record_custom_metrics([("CustomMetrics/OutsideTransaction/Count", 1)], application=app)
